// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file Assembler.h */

/**
 * \defgroup Assembler Assembler module
 *
 * \brief
 * Contains the operator and assembler classes:
 * @{ <img src="assembler.png"> @}
 */

#ifndef AMDIS_ASSEMBLER_H
#define AMDIS_ASSEMBLER_H

#include <vector>
#include "FixVec.h"
#include "MemoryManager.h"
#include "Operator.h"

namespace AMDiS {

  class ElInfo;
  class Element;
  class Assembler;
  class Quadrature;
  class FastQuadrature;
  class FiniteElemSpace;
  class ElementMatrix;
  class ElementVector;
  class BasisFunction;
  class Q00PsiPhi;
  class Q10PsiPhi;
  class Q01PsiPhi;
  class Q11PsiPhi;
  class Q0Psi;
  class Q1Psi;
  class Q2Psi;
  // class Operator;
  // class OperatorTerm;

  template<typename T> class DOFVectorBase;

  // enum FirstOrderType;

  // ============================================================================
  // ===== class SubAssembler ===================================================
  // ============================================================================

  /**
   * \ingroup Assembler
   * 
   * \brief
   * Base class for SecondOrderAssembler, FirstOrderAssembler, 
   * ZeroOrderAssembler. The task of a SubAssembler is to assemble a list of 
   * terms of a spezial order and add their contributions to a DOFMatrix or a 
   * DOFVector. An Assembler can consist of up to four SubAssemblers: one 
   * SecondOrderAssembler for second order terms, one ZeroOrderAssembler for 
   * terms of order zero, and two FirstOrderAssemblers. One for terms with
   * derivatives of the basis functions connected to to row DOFs and one for 
   * those connected to the column DOFs.
   */
  class SubAssembler
  {
  public:
    MEMORY_MANAGED(SubAssembler);

    /** \brief
     * Creates a SubAssembler belonging to assembler for the terms of given 
     * order of Operator op. If order is equal to one, type spezifies what kind 
     * of FirstOrderType are to assemble. During construction of a SubAssembler
     * the needs and properties of the terms are considered.
     */
    SubAssembler(Operator *op,
		 Assembler *assembler,
		 Quadrature *quadrat,
		 int order, 
		 bool optimized,
		 FirstOrderType type = GRD_PHI);

    /** \brief
     * Destructor
     */
    virtual ~SubAssembler() {};

    /** \brief
     * Calculates the element matrix for elInfo and adds it to mat. Memory
     * for mat must be provided by the caller.
     */
    virtual void calculateElementMatrix(const ElInfo *elInfo,
					ElementMatrix *mat) = 0;

    /** \brief
     * Calculates the element vector for elInfo and adds it to vec. Memory
     * for vec must be provided by the caller.
     */
    virtual void calculateElementVector(const ElInfo *elInfo,
					ElementVector *vec) = 0;

    /** \brief
     * Returns \ref terms
     */
    inline ::std::vector<OperatorTerm*> *getTerms() { 
      return &terms; 
    };

    /** \brief
     * Returns \ref quadrature.
     */ 
    inline Quadrature *getQuadrature() {
      return quadrature;
    };

    /** \brief
     * Sets \ref quadrature to q.
     */
    inline void setQuadrature(Quadrature* q) {
      /*     TEST_EXIT(!quadrature)("quadrature already set\n"); */
      quadrature = q;
    };
  
    /** \brief
     * Returns a vector with the world coordinates of the quadrature points
     * of \ref quadrature on the element of elInfo. 
     * Used by \ref OperatorTerm::initElement().
     */
    WorldVector<double>* getCoordsAtQPs(const ElInfo* elInfo, 
					Quadrature *quad = NULL);

    /** \brief
     * DOFVector dv evaluated at quadrature points.
     * Used by \ref OperatorTerm::initElement().
     */
    double* getVectorAtQPs(DOFVectorBase<double>* dv, const ElInfo* elInfo,
			   Quadrature *quad = NULL);

    /** \brief
     * Gradients of DOFVector dv evaluated at quadrature points.
     * Used by \ref OperatorTerm::initElement().
     */
    WorldVector<double>* getGradientsAtQPs(DOFVectorBase<double>* dv, 
					   const ElInfo* elInfo,
					   Quadrature *quad = NULL);

    /** \brief
     * Called once for each ElInfo when \ref calculateElementMatrix() or
     * \ref calculateElementVector() is called for the first time for this
     * Element.
     */
    virtual void initElement(const ElInfo *elInfo,
			     Quadrature *quad = NULL);

    /** \brief
     * Returns \ref psiFast.
     */
    const FastQuadrature *getPsiFast() const { 
      return psiFast; 
    };

    /** \brief
     * Returns \ref phiFast.
     */
    const FastQuadrature *getPhiFast() const { 
      return phiFast; 
    };

  protected:
    /** \brief
     * Updates \ref psiFast and \ref phiFast.
     */
    FastQuadrature *updateFastQuadrature(FastQuadrature *quadFast,
					 const BasisFunction *psi,
					 Flag updateFlag);
  
  protected:
    /** \brief
     * Problem dimension
     */
    int dim;

    /** \brief
     * Number of rows of the element matrix and length of the element
     * vector. Is equal to the number of row basis functions
     */
    int nRow;

    /** \brief
     * Number of columns of the element matrix. Is equal to the number
     * of column basis functions
     */
    int nCol;

    /** \brief
     * Used for \ref getVectorAtQPs().
     */
    class ValuesAtQPs {
    public:
      Vector<double> values;
      bool valid;
    };

    /** \brief
     * Used for \ref getGradientsAtQPs().
     */
    class GradientsAtQPs {
    public:
      Vector<WorldVector<double> > values;
      bool valid;
    };

    /** \brief
     * Used for \ref getVectorAtQPs().
     */
    ::std::map<const DOFVectorBase<double>*, ValuesAtQPs*> valuesAtQPs;

    /** \brief
     * Used for \ref getGradientsAtQPs().
     */
    ::std::map<const DOFVectorBase<double>*, GradientsAtQPs*> gradientsAtQPs;

    /** \brief
     * Set and updated by \ref initElement() for each ElInfo. 
     * coordsAtQPs[i] points to the coordinates of the i-th quadrature point.
     */
    WorldVector<double> *coordsAtQPs;

    /** \brief
     * Used for \ref getCoordsAtQPs().
     */
    bool coordsValid;

    /** \brief
     * Needed Quadrature. Constructed in the constructor of SubAssembler
     */
    Quadrature *quadrature;

    /** \brief
     * FastQuadrature for row basis functions
     */
    FastQuadrature *psiFast;

    /** \brief
     * FastQuadrature for column basis functions
     */
    FastQuadrature *phiFast;

    /** \brief
     * Corresponding Assembler
     */
    Assembler* owner;

    /** \brief
     * Flag that specifies whether the element matrix is symmetric.
     */
    bool symmetric;

    /** \brief
     * List of all terms with a contribution to this SubAssembler
     */
    ::std::vector<OperatorTerm*> terms;

    /** \brief
     *
     */
    bool opt;

    /** \brief
     *
     */
    bool firstCall;

    friend class Assembler;
  };

  // ============================================================================
  // ===== class ZeroOrderAssembler =============================================
  // ============================================================================

  /**
   * \ingroup Assembler
   * 
   * \brief
   * SubAssembler for zero order terms.
   */
  class ZeroOrderAssembler : public SubAssembler
  {
  public:
    MEMORY_MANAGED(ZeroOrderAssembler);

    /** \brief
     * Creates and returns the ZeroOrderAssembler for Operator op and
     * the given assembler. If all terms are piecewise constant precalculated 
     * integrals can be used while assembling and the returned 
     * ZeroOrderAssembler is of type Pre0. Otherwise a Quad0 object will
     * be returned.
     */
    static ZeroOrderAssembler* getSubAssembler(Operator *op,
					       Assembler *assembler,
					       Quadrature *quadrat,
					       bool optimized);


    /** \brief
     * Destructor.
     */
    virtual ~ZeroOrderAssembler() {};

  protected:
    /** \brief
     * Constructor.
     */
    ZeroOrderAssembler(Operator *op, 
		       Assembler *assembler, 
		       Quadrature *quadrat,
		       bool optimized);

  protected:
    /** \brief
     * List of all yet created optimized SubAssembler objects.
     */
    static ::std::vector<SubAssembler*> optimizedSubAssemblers;

    /** \brief
     * List of all yet created standard SubAssembler objects.
     */ 
    static ::std::vector<SubAssembler*> standardSubAssemblers;
  };

  // ============================================================================
  // ===== class Stand0 =========================================================
  // ============================================================================

  /**
   * \ingroup Assembler
   * 
   * \brief
   * Standard zero order assembler.
   */
  class Stand0 :  public ZeroOrderAssembler 
  {
  public:
    MEMORY_MANAGED(Stand0);

    /** \brief
     * Constructor.
     */
    Stand0(Operator *op, Assembler *assembler, Quadrature *quad);

    /** \brief
     * Implements SubAssembler::calculateElementMatrix().
     */
    void calculateElementMatrix(const ElInfo *elInfo, ElementMatrix *mat);

    /** \brief
     * Implements SubAssembler::calculateElementVector().
     */
    void calculateElementVector(const ElInfo *elInfo, ElementVector *vec);
  };

  // ============================================================================
  // ===== class Quad0 ==========================================================
  // ============================================================================

  /** 
   * \ingroup Assembler
   *
   * \brief
   * Zero order assembler using fast quadratures.
   */
  class Quad0 :  public ZeroOrderAssembler 
  {
  public:
    MEMORY_MANAGED(Quad0);

    /** \brief
     * Constructor.
     */
    Quad0(Operator *op, Assembler *assembler, Quadrature *quad);

    /** \brief
     * Implements SubAssembler::calculateElementMatrix().
     */
    void calculateElementMatrix(const ElInfo *elInfo, ElementMatrix *mat);

    /** \brief
     * Implements SubAssembler::calculateElementVector().
     */
    void calculateElementVector(const ElInfo *elInfo, ElementVector *vec);
  };


  // ============================================================================
  // ===== class Pre0 ===========================================================
  // ============================================================================

  /**
   * \ingroup Assembler
   * 
   * \brief
   * Zero order assembler using precaculated integrals.
   */
  class Pre0 : public ZeroOrderAssembler
  {
  public:
    MEMORY_MANAGED(Pre0);

    /** \brief
     * Constructor.
     */
    Pre0(Operator *op, Assembler *assembler, Quadrature *quad);

    /** \brief
     * Implements SubAssembler::calculateElementMatrix().
     */
    void calculateElementMatrix(const ElInfo *elInfo, ElementMatrix *mat);

    /** \brief
     * Implements SubAssembler::calculateElementVector().
     */
    void calculateElementVector(const ElInfo *elInfo, ElementVector *vec);

  protected:
    /** \brief
     * Integral of the product of psi and phi.
     */
    const Q00PsiPhi *q00;

    /** \brief
     * Integral of psi.
     */
    const Q0Psi *q0;
 
    friend class ZeroOrderAssembler;
  };

  // ============================================================================
  // ===== class FirstOrderAssembler ============================================
  // ============================================================================

  /**
   * \ingroup Assembler
   * 
   * \brief
   * SubAssembler for first order terms.
   */
  class FirstOrderAssembler : public SubAssembler
  {
  public:
    MEMORY_MANAGED(FirstOrderAssembler);

    /** \brief
     * Creates and returns the FirstOrderAssembler for Operator op and
     * the given assembler. If all terms are piecewise constant precalculated 
     * integrals can be used while assembling and the returned 
     * ZeroOrderAssembler is of type Pre0. Otherwise a Quad0 object will
     * be returned.
     */
    static FirstOrderAssembler* getSubAssembler(Operator *op,
						Assembler *assembler,
						Quadrature *quadrat,
						FirstOrderType type,
						bool optimized);
  
    /** \brief
     * Destructor.
     */
    virtual ~FirstOrderAssembler() {};

  protected:
    /** \brief
     * Constructor.
     */
    FirstOrderAssembler(Operator *op,
			Assembler *assembler,
			Quadrature *quadrat,
			bool optimized,
			FirstOrderType type);


  protected:
    /** \brief
     * List of all yet created optimized zero order assemblers for grdPsi.
     */
    static ::std::vector<SubAssembler*> optimizedSubAssemblersGrdPsi;

    /** \brief
     * List of all yet created standard zero order assemblers for grdPsi.
     */
    static ::std::vector<SubAssembler*> standardSubAssemblersGrdPsi;

    /** \brief
     * List of all yet created optimized zero order assemblers for grdPhi.
     */
    static ::std::vector<SubAssembler*> optimizedSubAssemblersGrdPhi;

    /** \brief
     * List of all yet created standard zero order assemblers for grdPhi.
     */
    static ::std::vector<SubAssembler*> standardSubAssemblersGrdPhi;
  };

  // ============================================================================
  // ===== class Stand10 ========================================================
  // ============================================================================

  /**
   * \ingroup Assembler
   * 
   * \brief
   * Standard first order assembler for grdPsi.
   */
  class Stand10 : public FirstOrderAssembler
  {
  public:
    MEMORY_MANAGED(Stand10);

    /** \brief
     * Constructor.
     */
    Stand10(Operator *op, Assembler *assembler, Quadrature *quad);

    /** \brief
     * Implements SubAssembler::calculateElementMatrix().
     */
    void calculateElementMatrix(const ElInfo *elInfo, ElementMatrix *mat);

    /** \brief
     * Implements SubAssembler::calculateElementVector().
     */
    void calculateElementVector(const ElInfo *, ElementVector */*vec*/);
  };

  // ============================================================================
  // ===== class Stand10 ========================================================
  // ============================================================================

  /**
   * \ingroup Assembler
   * 
   * \brief
   * Standard first order assembler for grdPhi.
   */
  class Stand01 : public FirstOrderAssembler
  {
  public:
    MEMORY_MANAGED(Stand01);

    /** \brief
     * Constructor.
     */
    Stand01(Operator *op, Assembler *assembler, Quadrature *quad);

    /** \brief
     * Implements SubAssembler::calculateElementMatrix().
     */
    void calculateElementMatrix(const ElInfo *elInfo, ElementMatrix *mat);

    /** \brief
     * Implements SubAssembler::calculateElementVector().
     */
    void calculateElementVector(const ElInfo *, ElementVector *) {
      ERROR_EXIT("should not be called\n");
    };
  };

  // ============================================================================
  // ===== class Quad10 =========================================================
  // ============================================================================

  /** \brief
   * First order assembler for grdPsi using fast quadratures.
   */
  class Quad10 : public FirstOrderAssembler
  {
  public:
    MEMORY_MANAGED(Quad10);

    /** \brief
     * Constructor.
     */
    Quad10(Operator *op, Assembler *assembler, Quadrature *quad);

    /** \brief
     * Implements SubAssembler::calculateElementMatrix().
     */
    void calculateElementMatrix(const ElInfo *elInfo, ElementMatrix *mat);

    /** \brief
     * Implements SubAssembler::calculateElementVector().
     */
    void calculateElementVector(const ElInfo *, ElementVector */*vec*/);
  };

  // ============================================================================
  // ===== class Quad01 =========================================================
  // ============================================================================

  /** \brief
   * First order assembler for grdPhi using fast quadratures.
   */
  class Quad01 : public FirstOrderAssembler 
  {
  public:
    MEMORY_MANAGED(Quad01);

    /** \brief
     * Constructor.
     */
    Quad01(Operator *op, Assembler *assembler, Quadrature *quad);

    /** \brief
     * Implements SubAssembler::calculateElementMatrix().
     */
    void calculateElementMatrix(const ElInfo *elInfo, ElementMatrix *mat);

    /** \brief
     * Implements SubAssembler::calculateElementVector().
     */
    void calculateElementVector(const ElInfo *, ElementVector */*vec*/) {
      ERROR_EXIT("should not be called\n");
    };
  };

  // ============================================================================
  // ===== class Pre10 ==========================================================
  // ============================================================================

  /** \brief
   * First order assembler for grdPsi using precalculated integrals
   */
  class Pre10 : public FirstOrderAssembler 
  {
  public:
    MEMORY_MANAGED(Pre10);

    /** \brief
     * Constructor.
     */
    Pre10(Operator *op, Assembler *assembler, Quadrature *quad);

    /** \brief
     * Implements SubAssembler::calculateElementMatrix().
     */
    void calculateElementMatrix(const ElInfo *elInfo, ElementMatrix *mat);

    /** \brief
     * Implements SubAssembler::calculateElementVector().
     */
    void calculateElementVector(const ElInfo *, ElementVector */*vec*/);

  protected:
    /** \brief
     * Integral of the product of the derivative of psi and phi.
     */
    const Q10PsiPhi *q10;

    /** \brief
     * Integral of the derivative of psi.
     */
    const Q1Psi *q1;

    friend class FirstOrderAssembler;
  };

  // ============================================================================
  // ===== class Pre01 ==========================================================
  // ============================================================================

  /** \brief
   *  First order assembler for grdPhi using precalculated integrals
   */
  class Pre01 : public FirstOrderAssembler 
  {
  public:
    MEMORY_MANAGED(Pre01);

    /** \brief
     * Constructor.
     */
    Pre01(Operator *op, Assembler *assembler, Quadrature *quad);

    /** \brief
     * Implements SubAssembler::calculateElementMatrix().
     */
    void calculateElementMatrix(const ElInfo *elInfo, ElementMatrix *mat);

    /** \brief
     * Implements SubAssembler::calculateElementVector().
     */
    void calculateElementVector(const ElInfo *, ElementVector */*vec*/) {
      ERROR_EXIT("should not be called\n");
    };

  protected:
    /** \brief
     * Integral of the product of psi and the derivative of phi.
     */
    const Q01PsiPhi *q01;

    /** \brief
     * Integral of the derivative of phi.
     */
    const Q1Psi *q1;

    friend class FirstOrderAssembler;
  };

  // ============================================================================
  // ===== class SecondOrderAssembler ===========================================
  // ============================================================================

  /**
   * \ingroup Assembler
   * 
   * \brief
   * SubAssembler for second order terms.
   */
  class SecondOrderAssembler : public SubAssembler
  {
  public:
    MEMORY_MANAGED(SecondOrderAssembler);

    /** \brief
     * Creates and returns the SecondOrderAssembler for Operator op and
     * the given assembler. If all terms are piecewise constant precalculated 
     * integrals can be used while assembling and the returned 
     * ZeroOrderAssembler is of type Pre0. Otherwise a Quad0 object will
     * be returned.
     */
    static SecondOrderAssembler* getSubAssembler(Operator *op,
						 Assembler *assembler,
						 Quadrature *quadrat,
						 bool optimized);

    /** \brief
     * Destructor.
     */
    virtual ~SecondOrderAssembler() {};

  protected:
    /** \brief
     * Constructor.
     */
    SecondOrderAssembler(Operator *op,
			 Assembler *assembler,
			 Quadrature *quadrat,
			 bool optimized);

  protected:
    /** \brief
     * List of all yet created optimized second order assemblers.
     */
    static ::std::vector<SubAssembler*> optimizedSubAssemblers;

    /** \brief
     * List of all yet created standard second order assemblers.
     */
    static ::std::vector<SubAssembler*> standardSubAssemblers;
  };

  // ============================================================================
  // ===== class Stand2 =========================================================
  // ============================================================================

  /**
   * \ingroup Assembler
   * 
   * \brief
   * Standard second order assembler
   */
  class Stand2 : public SecondOrderAssembler 
  {
  public:
    MEMORY_MANAGED(Stand2);

    /** \brief
     * Constructor.
     */
    Stand2(Operator *op, Assembler *assembler, Quadrature *quad);

    /** \brief
     * Implements SubAssembler::calculateElementMatrix().
     */
    void calculateElementMatrix(const ElInfo *elInfo, ElementMatrix *mat);

    /** \brief
     * Implements SubAssembler::calculateElementVector().
     */
    void calculateElementVector(const ElInfo *, ElementVector */*vec*/) {
      ERROR_EXIT("should not be called\n");
    };
  };

  // ============================================================================
  // ===== class Quad2 ==========================================================
  // ============================================================================

  /** \brief
   * Second order assembler using fast quadratures.
   */
  class Quad2 : public SecondOrderAssembler 
  {
  public:
    MEMORY_MANAGED(Quad2);

    /** \brief
     * Constructor.
     */
    Quad2(Operator *op, Assembler *assembler, Quadrature *quad);

    /** \brief
     * Implements SubAssembler::calculateElementMatrix().
     */
    void calculateElementMatrix(const ElInfo *elInfo, ElementMatrix *mat);

    /** \brief
     * Implements SubAssembler::calculateElementVector().
     */
    void calculateElementVector(const ElInfo *, ElementVector */*vec*/) {
      ERROR_EXIT("should not be called\n");
    };
  };

  // ============================================================================
  // ===== class Pre2 ===========================================================
  // ============================================================================

  /** \brief
   * Second order assembler using predefined integrals.
   */
  class Pre2 : public SecondOrderAssembler 
  {
  public:
    MEMORY_MANAGED(Pre2);

    /** \brief
     * Constructor.
     */
    Pre2(Operator *op, Assembler *assembler, Quadrature *quad);

    /** \brief
     * Implements SubAssembler::calculateElementMatrix().
     */
    void calculateElementMatrix(const ElInfo *elInfo, ElementMatrix *mat);

    /** \brief
     * Implements SubAssembler::calculateElementVector().
     */
    void calculateElementVector(const ElInfo *, ElementVector */*vec*/) {
      ERROR_EXIT("should not be called\n");
    };

  protected:
    /** \brief
     * Integral of the product of the derivative of psi and the derivative 
     * of phi.
     */
    const Q11PsiPhi *q11;

    friend class SecondOrderAssembler;
  };

  // ============================================================================
  // ===== class Assembler ======================================================
  // ============================================================================

  /**
   * \ingroup Assembler
   * 
   * \brief
   * Assembles element matrices and vectors for a given Operator. Uses
   * one SubAssembler for all second order terms of the Operator, one for all
   * first order terms, and one for all zero order terms.
   */
  class Assembler
  {
  public:
    MEMORY_MANAGED(Assembler);

    /** \brief
     * Constructor.
     */
    Assembler(Operator  *op,
	      const FiniteElemSpace *rowFESpace_,
	      const FiniteElemSpace *colFESpace_=NULL);

    virtual ~Assembler() {};

    ElementMatrix *initElementMatrix(ElementMatrix *elMat, 
				     const ElInfo *elInfo);


    ElementVector *initElementVector(ElementVector *elVec,
				     const ElInfo *elInfo);

    /** \brief
     * Assembles the element matrix for the given ElInfo
     */
    virtual void calculateElementMatrix(const ElInfo *elInfo, 
					ElementMatrix *userMat, 
					double factor=1.0);

    /** \brief
     * Assembles the element vector for the given ElInfo
     */
    virtual void calculateElementVector(const ElInfo *elInfo, 
					ElementVector *userVec, 
					double factor=1.0);

    /** \brief
     * Returns \ref rowFESpace.
     */
    inline const FiniteElemSpace* getRowFESpace() { 
      return rowFESpace; 
    };

    /** \brief
     * Returns \ref colFESpace.
     */
    inline const FiniteElemSpace* getColFESpace() { 
      return colFESpace; 
    };

    /** \brief
     * Returns \ref nRow.
     */
    inline int getNRow() { 
      return nRow; 
    };

    /** \brief
     * Returns \ref nCol.
     */
    inline int getNCol() { 
      return nCol; 
    };

    /** \brief
     * Sets \ref rememberElMat.
     */
    inline void rememberElementMatrix(bool rem) { 
      rememberElMat = rem; 
    };

    /** \brief
     * Sets \ref rememberElVec.
     */
    inline void rememberElementVector(bool rem) { 
      rememberElVec = rem; 
    };

    /** \brief
     * Returns \ref zeroOrderAssembler.
     */
    inline ZeroOrderAssembler* getZeroOrderAssembler() {
      return zeroOrderAssembler;
    };

    /** \brief
     * Returns \ref firstOrderAssemblerGrdPsi or \ref firstOrderAssemblerGrdPhi
     * depending on type.
     */
    inline FirstOrderAssembler* getFirstOrderAssembler(FirstOrderType type = 
						       GRD_PSI) 
    {
      return
	(type == GRD_PSI) ? 
	firstOrderAssemblerGrdPsi : 
	firstOrderAssemblerGrdPhi;
    };

    /** \brief
     * Returns \ref secondOrderAssembler.
     */
    inline SecondOrderAssembler* getSecondOrderAssembler() {
      return secondOrderAssembler;
    };

    /** \brief
     * Returns \ref operat;
     */
    inline Operator* getOperator() { 
      return operat; 
    };

    /** \brief
     * Initialisation for the given ElInfo. The call is deligated to
     * the sub assemblers.
     */
    void initElement(const ElInfo *elInfo,
		     Quadrature *quad = NULL);

    /** \brief
     * Sets quadratures of all sub assemblers.
     */
    void setQuadratures(Quadrature *quad2,
			Quadrature *quad1GrdPsi,
			Quadrature *quad1GrdPhi,
			Quadrature *quad0) 
    {
      if(secondOrderAssembler) {
	TEST_EXIT(!secondOrderAssembler->getQuadrature())
	  ("quadrature already existing\n");
	secondOrderAssembler->setQuadrature(quad2);
      }
      if(firstOrderAssemblerGrdPsi) {
	TEST_EXIT(!firstOrderAssemblerGrdPsi->getQuadrature())
	  ("quadrature already existing\n");
	firstOrderAssemblerGrdPsi->setQuadrature(quad1GrdPsi);
      }
      if(firstOrderAssemblerGrdPhi) {
	TEST_EXIT(!firstOrderAssemblerGrdPhi->getQuadrature())
	  ("quadrature already existing\n");
	firstOrderAssemblerGrdPhi->setQuadrature(quad1GrdPhi);
      }
      if(zeroOrderAssembler) {
	TEST_EXIT(!zeroOrderAssembler->getQuadrature())
	  ("quadrature already existing\n");
	zeroOrderAssembler->setQuadrature(quad0);
      }
    };

  protected:
    /** \brief
     * Vector assembling by element matrix-vector multiplication. 
     * Usefull if an element matrix was already calculated.
     */
    void matVecAssemble(const ElInfo *elInfo, ElementVector *vec);

    /** \brief
     * Checks whether this is a new travese.
     */
    inline void checkForNewTraverse() {
      if (lastTraverseId != ElInfo::traverseId) {
	lastVecEl = lastMatEl = NULL;
	lastTraverseId = ElInfo::traverseId;
      }
    };

    /** \brief
     * Checks whether quadratures for sub assemblers are already set.
     * If not they will be created.
     */
    void checkQuadratures();

  protected:
    /** \brief
     * Operator this Assembler belongs to.
     */
    Operator *operat;

    /** \brief
     * Row FiniteElemSpace.
     */
    const FiniteElemSpace *rowFESpace;

    /** \brief
     * Column FiniteElemSpace.
     */
    const FiniteElemSpace *colFESpace;

    /** \brief
     * Number of rows.
     */
    int nRow;

    /** \brief
     * Number of columns.
     */
    int nCol;

    /** \brief
     * SubAssembler for the second order terms
     */
    SecondOrderAssembler *secondOrderAssembler;

    /** \brief
     * SubAssembler for the first order terms (grdPsi)
     */
    FirstOrderAssembler  *firstOrderAssemblerGrdPsi;

    /** \brief
     * SubAssembler for the first order terms (grdPhi)
     */
    FirstOrderAssembler  *firstOrderAssemblerGrdPhi;  

    /** \brief
     * SubAssembler for the zero order terms
     */
    ZeroOrderAssembler   *zeroOrderAssembler;

    bool remember;

    /** \brief
     * Determines whether the element matrix should be stored locally.
     */
    bool rememberElMat;

    /** \brief
     * Determines whether the element vector should be stored locally.
     */
    bool rememberElVec;

    /** \brief
     * locally stored element matrix
     */
    ElementMatrix* elementMatrix;

    /** \brief
     * locally stored element vector
     */
    ElementVector* elementVector;
  
    /** \brief
     * Used to check whether \ref initElement() must be called, because
     * a new Element is visited.
     */
    Element* lastMatEl;

    /** \brief
     * Used to check whether \ref initElement() must be called, because
     * a new Element is visited.
     */
    Element* lastVecEl;

    /** \brief
     * Used to check for new traverse.
     */
    int lastTraverseId;

    friend class SubAssembler;
    friend class ZeroOrderAssembler;
    friend class FirstOrderAssembler;
    friend class SecondOrderAssembler;
  };

  // ============================================================================
  // ===== class StandardAssembler =============================================
  // ============================================================================

  /**
   * \ingroup Assembler
   *
   * \brief
   * Assembler using non optimized sub assemblers.
   */
  class StandardAssembler : public Assembler
  {
  public:
    MEMORY_MANAGED(StandardAssembler);

    /** \brief
     * Constructor.
     */
    StandardAssembler(Operator  *op,
		      Quadrature *quad2,
		      Quadrature *quad1GrdPsi,
		      Quadrature *quad1GrdPhi,
		      Quadrature *quad0,
		      const FiniteElemSpace *rowFESpace_,
		      const FiniteElemSpace *colFESpace_=NULL);
  };

  // ============================================================================
  // ===== class OptimizedAssembler =============================================
  // ============================================================================

  /**
   * \ingroup Assembler
   *
   * \brief
   * Assembler using optimized sub assemblers.
   */
  class OptimizedAssembler : public Assembler
  {
  public:
    MEMORY_MANAGED(OptimizedAssembler);

    /** \brief
     * Constructor.
     */
    OptimizedAssembler(Operator  *op,
		       Quadrature *quad2,
		       Quadrature *quad1GrdPsi,
		       Quadrature *quad1GrdPhi,
		       Quadrature *quad0,
		       const FiniteElemSpace *rowFESpace_,
		       const FiniteElemSpace *colFESpace_=NULL);
  };

}

#endif // AMDIS_ASSEMBLER_H
