#include "DOFVector.h"
#include "Traverse.h"
#include "DualTraverse.h"
#include "FixVec.h"

namespace AMDiS {

  template<>
  void DOFVector<double>::coarseRestrict(RCNeighbourList& list, int n)
  {
    switch(coarsenOperation) {
    case NO_OPERATION:
      return;
      break;
    case COARSE_RESTRICT:
      (const_cast<BasisFunction*>(feSpace->getBasisFcts()))->coarseRestr(this, &list, n);
      break;
    case COARSE_INTERPOL:
      (const_cast<BasisFunction*>(feSpace->getBasisFcts()))->coarseInter(this, &list, n);
      break;
    default:
      ERROR_EXIT("invalid coarsen operation\n");
    }
  }

  template<>
  void DOFVector<double>::refineInterpol(RCNeighbourList& list, int n)
  {
    (const_cast<BasisFunction*>(feSpace->getBasisFcts()))->refineInter(this, &list, n);
  }

  template<>
  void DOFVector<WorldVector<double> >::refineInterpol(RCNeighbourList& list, int n)
  {
    if(n < 1) return;
    int dim = feSpace->getMesh()->getDim();
    Element *el = list.getElement(0);
    int n0 = feSpace->getAdmin()->getNumberOfPreDOFs(VERTEX);
    DegreeOfFreedom dof0 = el->getDOF(0, n0);
    DegreeOfFreedom dof1 = el->getDOF(1, n0);
    DegreeOfFreedom dof_new = el->getChild(0)->getDOF(dim, n0);
    vec[dof_new] = vec[dof0];
    vec[dof_new] += vec[dof1];
    vec[dof_new] *= 0.5;
  }

  template<>
  DOFVector<WorldVector<double> >*
  DOFVector<double>::getGradient(DOFVector<WorldVector<double> > *grad) const 
  {
    FUNCNAME("DOFVector<double>::getGradient()");

    // define result vector
    static DOFVector<WorldVector<double> > *result = NULL;

    if(grad) {
      result = grad;
    } else {
      if(result && result->getFESpace() != feSpace) {
	DELETE result;
	result = NEW DOFVector<WorldVector<double> >(feSpace, "gradient");
      }
    }

    int i, j;

    Mesh *mesh = feSpace->getMesh();

    int dim = mesh->getDim();

    const BasisFunction *basFcts = feSpace->getBasisFcts();
    int nBasFcts = basFcts->getNumber();

    DOFAdmin *admin = feSpace->getAdmin();

    // count number of nodes and dofs per node
    ::std::vector<int> numNodeDOFs;
    ::std::vector<int> numNodePreDOFs;
    ::std::vector<DimVec<double>*> bary;

    int numNodes = 0;
    int numDOFs = 0;

    for(i = 0; i < dim + 1; i++) {
      GeoIndex geoIndex = INDEX_OF_DIM(i, dim);
      int numPositionNodes = mesh->getGeo(geoIndex);
      int numPreDOFs = admin->getNumberOfPreDOFs(i);
      for(j = 0; j < numPositionNodes; j++) {
	int dofs = basFcts->getNumberOfDOFs(geoIndex);
	numNodeDOFs.push_back(dofs);
	numDOFs += dofs;
	numNodePreDOFs.push_back(numPreDOFs);
      }
      numNodes += numPositionNodes;
    }

    TEST_EXIT(numNodes == mesh->getNumberOfNodes())
      ("invalid number of nodes\n");

    TEST_EXIT(numDOFs == nBasFcts)
      ("number of dofs != number of basis functions\n");

    for(i = 0; i < numDOFs; i++) {
      bary.push_back(basFcts->getCoords(i));
    }

    // traverse mesh
    ::std::vector<bool> visited(getUsedSize(), false);

    TraverseStack stack;

    Flag fillFlag = Mesh::CALL_LEAF_EL | Mesh::FILL_GRD_LAMBDA;

    ElInfo *elInfo = stack.traverseFirst(mesh, -1, fillFlag);

    while(elInfo) {

      const DegreeOfFreedom **dof = elInfo->getElement()->getDOF();

      const double *localUh = getLocalVector(elInfo->getElement(), NULL);

      const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();

      int localDOFNr = 0;
      for(i = 0; i < numNodes; i++) { // for all nodes
	for(j = 0; j < numNodeDOFs[i]; j++) { // for all dofs at this node
	  DegreeOfFreedom dofIndex = dof[i][numNodePreDOFs[localDOFNr] + j];
	  if(!visited[dofIndex]) {
	  
	    result[dofIndex] = basFcts->evalGrdUh(*(bary[localDOFNr]),
						  grdLambda, 
						  localUh, 
						  NULL);

	    visited[dofIndex] = true;
	  }
	  localDOFNr++;
	}
      }

      elInfo = stack.traverseNext(elInfo);
    }

    return result;
  }

  template<>
  DOFVector<WorldVector<double> >*
  DOFVector<double>::getRecoveryGradient(DOFVector<WorldVector<double> > *grad) const 
  {
    FUNCNAME("DOFVector<double>::getRecoveryGradient()");

    // define result vector
    static DOFVector<WorldVector<double> > *vec = NULL;

    DOFVector<WorldVector<double> > *result = grad;

    if(!result) {
      if(vec && vec->getFESpace() != feSpace) {
	DELETE vec;
	vec = NULL;
      }
      if(!vec) {
	vec = NEW DOFVector<WorldVector<double> >(feSpace, "gradient");
      }
      result = vec;
    }

    result->set(WorldVector<double>(DEFAULT_VALUE, 0.0));

    DOFVector<double> volume(feSpace, "volume");
    volume.set(0.0);

    int i;

    Mesh *mesh = feSpace->getMesh();

    int dim = mesh->getDim();

    const BasisFunction *basFcts = feSpace->getBasisFcts();

    DOFAdmin *admin = feSpace->getAdmin();

    int numPreDOFs = admin->getNumberOfPreDOFs(0);

    DimVec<double> bary(dim, DEFAULT_VALUE, (1.0 / (dim + 1.0)));

    // traverse mesh
    TraverseStack stack;

    Flag fillFlag = 
      Mesh::CALL_LEAF_EL | Mesh::FILL_DET | Mesh::FILL_GRD_LAMBDA | Mesh::FILL_COORDS;

    ElInfo *elInfo = stack.traverseFirst(mesh, -1, fillFlag);

    while(elInfo) {
      double det = elInfo->getDet();
      const DegreeOfFreedom **dof = elInfo->getElement()->getDOF();
      const double *localUh = getLocalVector(elInfo->getElement(), NULL);
      const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();
      const WorldVector<double> &grd = basFcts->evalGrdUh(bary,
							  grdLambda, 
							  localUh, 
							  NULL);

      for(i = 0; i < dim + 1; i++) {
	DegreeOfFreedom dofIndex = dof[i][numPreDOFs];
	(*result)[dofIndex] += grd * det;
	volume[dofIndex] += det;
      }

      elInfo = stack.traverseNext(elInfo);
    }

    DOFVector<double>::Iterator volIt(&volume, USED_DOFS);
    DOFVector<WorldVector<double> >::Iterator grdIt(result, USED_DOFS);

    for(volIt.reset(), grdIt.reset(); !volIt.end(); ++volIt, ++grdIt) {
      if(*volIt != 0.0) {
	*grdIt *= 1.0/(*volIt);
      }
    }

    return result;
  }

  template<>
  const WorldVector<double> *DOFVectorBase<double>::getGrdAtQPs(const ElInfo         *elInfo, 
								const Quadrature     *quad,
								const FastQuadrature *quadFast,
								WorldVector<double>  *grdAtQPs) const
  {
    FUNCNAME("DOFVector<double>::getGrdAtQPs()");
  
    TEST_EXIT(quad || quadFast)("neither quad nor quadFast defined\n");

    if(quad && quadFast) {
      TEST_EXIT(quad == quadFast->getQuadrature())
	("quad != quadFast->quadrature\n");
    }

    TEST_EXIT(!quadFast || quadFast->getBasisFunctions() == feSpace->getBasisFcts())
      ("invalid basis functions");

    Element *el = elInfo->getElement(); 

    int dim = elInfo->getMesh()->getDim();
    int dow = Global::getGeo(WORLD);

    const Quadrature *quadrature = quadFast ? quadFast->getQuadrature() : quad;

    const BasisFunction *basFcts = feSpace->getBasisFcts();

    int numPoints = quadrature->getNumPoints();
    int nBasFcts  = basFcts->getNumber();
    int i, j, k, l;

    static WorldVector<double> *grd = NULL;

    WorldVector<double> *result;

    if (grdAtQPs) {
      result = grdAtQPs;
    } else {
      if(grd) delete [] grd;
      grd = new WorldVector<double>[numPoints];
      for(i = 0; i < numPoints; i++) {
	grd[i].set(0.0);
      }
      result = grd;
    }
  
    const double *localVec = getLocalVector(el, NULL);

    DimVec<double> grd1(dim, NO_INIT);
    int parts = Global::getGeo(PARTS, dim);
    const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();

    if(quadFast) {
      for (i = 0; i < numPoints; i++) {
	for(j = 0; j < dim + 1; j++)
	  grd1[j] = 0.0;

	for (j = 0; j < nBasFcts; j++) {
	  for (k = 0; k < parts; k++) {
	    grd1[k] += quadFast->getGradient(i, j, k) * localVec[j];
	  }
	}

	for(l=0; l < dow; l++) {
	  for (result[i][l] = k = 0; k < parts; k++) {
	    result[i][l] += grdLambda[k][l] * grd1[k];
	  }
	}
      }
    } else {
      DimVec<double> grdPhi(dim, NO_INIT);
      for (i = 0; i < numPoints; i++) {
	for(j = 0; j < dim + 1; j++)
	  grd1[j] = 0.0;
	for (j = 0; j < nBasFcts; j++) {
	  grdPhi = (*(basFcts->getGrdPhi(j)))(quad->getLambda(i));
	  for(k = 0; k < parts; k++) {
	    grd1[k] += grdPhi[k] * localVec[j];
	  }
	}

	for(l=0; l < dow; l++) {
	  for (result[i][l] = k = 0; k < parts; k++) {
	    result[i][l] += grdLambda[k][l] * grd1[k];
	  }
	}
      }
    }
  

    return const_cast<const WorldVector<double>*>(result);
  }

  template<>
  const WorldMatrix<double> *DOFVectorBase<double>::getD2AtQPs(const ElInfo         *elInfo, 
							       const Quadrature     *quad,
							       const FastQuadrature *quadFast,
							       WorldMatrix<double>  *d2AtQPs) const
  {
    FUNCNAME("DOFVector<double>::getD2AtQPs()");
  
    TEST_EXIT(quad || quadFast)("neither quad nor quadFast defined\n");

    if(quad && quadFast) {
      TEST_EXIT(quad == quadFast->getQuadrature())
	("quad != quadFast->quadrature\n");
    }

    TEST_EXIT(!quadFast || quadFast->getBasisFunctions() == feSpace->getBasisFcts())
      ("invalid basis functions");

    Element *el = elInfo->getElement(); 

    int dim = elInfo->getMesh()->getDim();
    int dow = Global::getGeo(WORLD);

    const Quadrature *quadrature = quadFast ? quadFast->getQuadrature() : quad;

    const BasisFunction *basFcts = feSpace->getBasisFcts();

    int numPoints = quadrature->getNumPoints();
    int nBasFcts  = basFcts->getNumber();
    int i, j, k, l, iq;

    static WorldMatrix<double> *vec = NULL;

    WorldMatrix<double> *result;

    if (d2AtQPs) {
      result = d2AtQPs;
    } else {
      if(vec) delete [] vec;
      vec = new WorldMatrix<double>[numPoints];
      for(i = 0; i < numPoints; i++) {
	vec[i].set(0.0);
      }
      result = vec;
    }
  
    const double *localVec = getLocalVector(el, NULL);

    DimMat<double> D2Tmp(dim, DEFAULT_VALUE, 0.0);
    int parts = Global::getGeo(PARTS, dim);
    const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();

    if(quadFast) {
      for (iq = 0; iq < numPoints; iq++) {
	for (k = 0; k < parts; k++)
	  for (l = 0; l < parts; l++)
	    D2Tmp[k][l] = 0.0;

	for (i = 0; i < nBasFcts; i++) {
	  for (k = 0; k < parts; k++)
	    for (l = 0; l < parts; l++)
	      D2Tmp[k][l] += localVec[i]* quadFast->getSecDer(iq, i, k, l);
	}

	for (i = 0; i < dow; i++)
	  for (j = 0; j < dow; j++) {
	    result[iq][i][j] = 0.0;
	    for (k = 0; k < parts; k++)
	      for (l = 0; l < parts; l++)
		result[iq][i][j] += grdLambda[k][i]*grdLambda[l][j]*D2Tmp[k][l];
	  }
      }
    } else {
      DimMat<double> D2Phi(dim, NO_INIT);
      for (iq = 0; iq < numPoints; iq++) {
	for (k = 0; k < parts; k++)
	  for (l = 0; l < parts; l++)
	    D2Tmp[k][l] = 0.0;

	for (i = 0; i < nBasFcts; i++) {
	  WARNING("not tested after index correction\n");
	  //D2Phi = (*(basFcts->getD2Phi(j)))(quad->getLambda(i));
	  D2Phi = (*(basFcts->getD2Phi(i)))(quad->getLambda(iq));
	  for (k = 0; k < parts; k++)
	    for (l = 0; l < parts; l++)
	      D2Tmp[k][l] += localVec[i] * D2Phi[k][l];
	}

	for (i = 0; i < dow; i++)
	  for (j = 0; j < dow; j++) {
	    result[iq][i][j] = 0.0;
	    for (k = 0; k < parts; k++)
	      for (l = 0; l < parts; l++)
		result[iq][i][j] += grdLambda[k][i]*grdLambda[l][j]*D2Tmp[k][l];
	  }
      }
    }
  

    return const_cast<const WorldMatrix<double>*>(result);
  }

  template<>
  void DOFVector<double>::interpol(DOFVector<double> *source, double factor) 
  {
    FUNCNAME("DOFVector<double>::interpol()");

    const FiniteElemSpace *sourceFeSpace = source->getFESpace();

    const BasisFunction *basisFcts = feSpace->getBasisFcts();
    const BasisFunction *sourceBasisFcts = sourceFeSpace->getBasisFcts();

    int nBasisFcts = basisFcts->getNumber();
    int nSourceBasisFcts = sourceBasisFcts->getNumber();

    this->set(0.0);

    DegreeOfFreedom *localIndices = GET_MEMORY(DegreeOfFreedom, nBasisFcts);
    double *sourceLocalCoeffs = GET_MEMORY(double, nSourceBasisFcts);

    if (feSpace->getMesh() == sourceFeSpace->getMesh()) {
      DimVec<double> *coords = NULL;
      TraverseStack stack;
      ElInfo *elInfo = stack.traverseFirst(feSpace->getMesh(), -1,
					   Mesh::CALL_LEAF_EL | 
					   Mesh::FILL_COORDS);

      while (elInfo) {
	Element *el = elInfo->getElement();

	basisFcts->getLocalIndices(el, feSpace->getAdmin(), localIndices);

	source->getLocalVector(el, sourceLocalCoeffs);

	for (int i = 0; i < nBasisFcts; i++) {
	  if (vec[localIndices[i]] == 0.0) {
	    coords = basisFcts->getCoords(i);
	    vec[localIndices[i]] = sourceBasisFcts->evalUh(*coords, sourceLocalCoeffs) * factor;
	  }
	}
	elInfo = stack.traverseNext(elInfo);
      }
    } else {
      DimVec<double> *coords1 = NULL;
      DimVec<double> coords2(feSpace->getMesh()->getDim(), NO_INIT);
      DualTraverse dualStack;
      ElInfo *elInfo1, *elInfo2;
      ElInfo *elInfoSmall, *elInfoLarge;
      WorldVector<double> worldVec;

      bool nextTraverse = dualStack.traverseFirst(feSpace->getMesh(),
						  sourceFeSpace->getMesh(),
						  -1, -1,
						  Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS,
						  Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS,
						  &elInfo1, &elInfo2,
						  &elInfoSmall, &elInfoLarge);
      while (nextTraverse) {     
	basisFcts->getLocalIndices(elInfo1->getElement(), 
				   feSpace->getAdmin(), 
				   localIndices);
	source->getLocalVector(elInfo2->getElement(), 
			       sourceLocalCoeffs);

	for (int i = 0; i < nBasisFcts; i++) {
	  if (vec[localIndices[i]] == 0.0) {
	    coords1 = basisFcts->getCoords(i);
	    elInfo1->coordToWorld(*coords1, &worldVec);
	    elInfo2->worldToCoord(worldVec, &coords2);
	  
	    bool isPositive = true;
	    for (int j = 0; j < coords2.size(); j++) {
	      if (coords2[j] < 0) {
		isPositive = false;
		break;
	      }
	    }
	  
	    if (isPositive) {
	      vec[localIndices[i]] = sourceBasisFcts->evalUh(coords2, sourceLocalCoeffs);	    
	    }	
	  }
	}
      
	nextTraverse = dualStack.traverseNext(&elInfo1, &elInfo2,
					      &elInfoSmall, &elInfoLarge);
      }
    }
  
    FREE_MEMORY(localIndices, DegreeOfFreedom, nBasisFcts);
    FREE_MEMORY(sourceLocalCoeffs, double, nSourceBasisFcts);
  }


  template<>
  void DOFVector<WorldVector<double> >::interpol(DOFVector<WorldVector<double> > *v, double factor) 
  {
    WorldVector<double> nul(DEFAULT_VALUE,0.0);

    this->set(nul);

    DimVec<double> *coords = NULL;

    const FiniteElemSpace *vFESpace = v->getFESpace();

    if (feSpace == vFESpace) {
      WARNING("same FE-spaces\n");
    }

    const BasisFunction *basFcts = feSpace->getBasisFcts();
    const BasisFunction *vBasFcts = vFESpace->getBasisFcts();

    int numBasFcts = basFcts->getNumber();
    int vNumBasFcts = vBasFcts->getNumber();

    if (feSpace->getMesh() == vFESpace->getMesh()) {
      DegreeOfFreedom *localIndices = GET_MEMORY(DegreeOfFreedom, numBasFcts);
      WorldVector<double> *vLocalCoeffs = NEW WorldVector<double>[vNumBasFcts];
      Mesh *mesh = feSpace->getMesh();
      TraverseStack stack;
      ElInfo *elInfo = stack.traverseFirst(mesh, -1,
					   Mesh::CALL_LEAF_EL | 
					   Mesh::FILL_COORDS);

      while (elInfo) {
	Element *el = elInfo->getElement();

	basFcts->getLocalIndices(el, feSpace->getAdmin(), localIndices);

	v->getLocalVector(el, vLocalCoeffs);

	for (int i = 0; i < numBasFcts; i++) {
	  if (vec[localIndices[i]] == nul) {
	    coords = basFcts->getCoords(i);
	    //	  for(j = 0; j < vNumBasFcts; j++) {
	    vec[localIndices[i]] += vBasFcts->evalUh(*coords, vLocalCoeffs,NULL) * factor;
	    //  }
	  }
	}
	elInfo = stack.traverseNext(elInfo);
      }

      FREE_MEMORY(localIndices, DegreeOfFreedom, numBasFcts);
      DELETE []   vLocalCoeffs;
    } else {
      ERROR_EXIT("not yet for dual traverse\n");
    }
  }


  template<>
  WorldVector<DOFVector<double>*> *DOFVector<double>::getGradient(WorldVector<DOFVector<double>*> *grad) const
  {
    FUNCNAME("DOFVector<double>::getGradient()");

    int i, j, k;

    Mesh *mesh = feSpace->getMesh();

    int dim = mesh->getDim();
    int dow = Global::getGeo(WORLD);

    const BasisFunction *basFcts = feSpace->getBasisFcts();
    int nBasFcts = basFcts->getNumber();

    DOFAdmin *admin = feSpace->getAdmin();

    // define result vector
    static WorldVector<DOFVector<double>*> *result = NULL;

    if (grad) {
      result = grad;
    } else {
      if (!result) {
	result = NEW WorldVector<DOFVector<double>*>;
	result->set(NULL);
      }
      for (i = 0; i < dow; i++) {
	if ((*result)[i] && (*result)[i]->getFESpace() != feSpace) {
	  DELETE (*result)[i];
	  (*result)[i] = NEW DOFVector<double>(feSpace, "gradient");
	}
      }
    }

    // count number of nodes and dofs per node
    ::std::vector<int> numNodeDOFs;
    ::std::vector<int> numNodePreDOFs;
    ::std::vector<DimVec<double>*> bary;

    int numNodes = 0;
    int numDOFs = 0;

    for (i = 0; i < dim + 1; i++) {
      GeoIndex geoIndex = INDEX_OF_DIM(i, dim);
      int numPositionNodes = mesh->getGeo(geoIndex);
      int numPreDOFs = admin->getNumberOfPreDOFs(i);
      for (j = 0; j < numPositionNodes; j++) {
	int dofs = basFcts->getNumberOfDOFs(geoIndex);
	numNodeDOFs.push_back(dofs);
	numDOFs += dofs;
	numNodePreDOFs.push_back(numPreDOFs);
      }
      numNodes += numPositionNodes;
    }

    TEST_EXIT(numNodes == mesh->getNumberOfNodes())
      ("invalid number of nodes\n");

    TEST_EXIT(numDOFs == nBasFcts)
      ("number of dofs != number of basis functions\n");

    for (i = 0; i < numDOFs; i++) {
      bary.push_back(basFcts->getCoords(i));
    }

    // traverse mesh
    ::std::vector<bool> visited(getUsedSize(), false);

    TraverseStack stack;

    Flag fillFlag = Mesh::CALL_LEAF_EL | Mesh::FILL_GRD_LAMBDA;

    ElInfo *elInfo = stack.traverseFirst(mesh, -1, fillFlag);

    WorldVector<double> grd;

    while (elInfo) {

      const DegreeOfFreedom **dof = elInfo->getElement()->getDOF();

      const double *localUh = getLocalVector(elInfo->getElement(), NULL);

      const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();

      int localDOFNr = 0;
      for (i = 0; i < numNodes; i++) { // for all nodes
	for (j = 0; j < numNodeDOFs[i]; j++) { // for all dofs at this node
	  DegreeOfFreedom dofIndex = dof[i][numNodePreDOFs[localDOFNr] + j];
	  if (!visited[dofIndex]) {
	  
	    grd = basFcts->evalGrdUh(*(bary[localDOFNr]),
				     grdLambda, 
				     localUh, 
				     NULL);

	    for(k = 0; k < dow; k++) {
	      (*result)[k][dofIndex] = grd[k];
	    }

	    visited[dofIndex] = true;
	  }
	  localDOFNr++;
	}
      }

      elInfo = stack.traverseNext(elInfo);
    }

    return result;
  }


  DOFVectorDOF::DOFVectorDOF() : DOFVector<DegreeOfFreedom>() {};


  void DOFVectorDOF::freeDOFContent(DegreeOfFreedom dof) {
    ::std::vector<DegreeOfFreedom>::iterator it;
    ::std::vector<DegreeOfFreedom>::iterator end = vec.end();
    DegreeOfFreedom pos;
    for(it = vec.begin(), pos = 0; it != end; ++it, ++pos) {
      if(*it == dof) *it = pos;
    }
  };


  WorldVector<DOFVector<double>*> *transform(DOFVector<WorldVector<double> > *vec,
					     WorldVector<DOFVector<double>*> *res)
  {
    FUNCNAME("transform()");

    TEST_EXIT(vec)("no vector\n");

    int i, j, dow = Global::getGeo(WORLD);

    static WorldVector<DOFVector<double>*> *result = NULL;

    if(!res && !result) {
      result = NEW WorldVector<DOFVector<double>*>;
      for(i = 0; i < dow; i++) {
	(*result)[i] = NEW DOFVector<double>(vec->getFESpace(), "transform");
      }
    }

    WorldVector<DOFVector<double>*> *r = res ? res : result;

    int vecSize = vec->getSize();
    for(i = 0; i < vecSize; i++) {
      for(j = 0; j < dow; j++) {
	(*((*r)[j]))[i] = (*vec)[i][j];
      }
    }

    return r;
  }

}

