#include "Element.h"
#include "DOFAdmin.h"
#include "Mesh.h"
#include "CoarseningManager.h"
#include "FixVec.h"
#include "ElementRegion_ED.h"

namespace AMDiS {


  int Element::getRegion() const 
  {
    ElementRegion_ED* red_;

    if (!elementData) return -1;
    red_=dynamic_cast<ElementRegion_ED*>(elementData->getElementData(ELEMENT_REGION));
    if (red_) 
      return red_->getRegion();
    
    return -1;
  }

  void Element::setDOFPtrs()
  {
    FUNCNAME("Element::setDOFPtrs");
    TEST_EXIT(mesh)("no mesh!\n");
    dof = mesh->createDOFPtrs();  
  }

  Element::Element(Mesh *aMesh)
  {
    mesh = aMesh;

    index = mesh ? mesh->getNextElementIndex() : -1; 
  
    child[0]  = NULL;
    child[1]  = NULL;

    dof       = mesh ? mesh->createDOFPtrs() : NULL;
    newCoord = NULL;

    elementData = NULL;
  }

  // call destructor through Mesh::freeElement !!!
  Element::~Element()
  {
    if(child[0]) DELETE child[0];
    if(child[1]) DELETE child[1];

    //if(elementData) DELETE elementData;

    if (newCoord)
      {
	DELETE newCoord;
      }
  }

  /****************************************************************************/
  /*  ATTENTION:                                                              */
  /*  new_dof_fct() destroys new_dof !!!!!!!!!!                               */
  /*  should be used only at the end of dof_compress()!!!!!                   */
  /****************************************************************************/

  /* CHANGE_DOFS_1 changes old dofs to NEGATIVE new dofs */

#define CHANGE_DOFS_1(el)						\
  ldof = el->dof[n0+i] + nd0;						\
    for (j = 0; j < nd; j++) {						\
      if ((k = ldof[j]) >= 0) {						\
	/* do it only once! (dofs are visited more than once) */	\
	ldof[j] = - admin->getMesh()->newDOF[k] - 1;			\
      } }

  /* CHANGE_DOFS_2 changes NEGATIVE new dofs to POSITIVE */

#define CHANGE_DOFS_2(el)						\
  ldof = el->dof[n0+i] + nd0;						\
    for (j = 0; j < nd; j++) {						\
      if ((k = ldof[j]) < 0) {						\
	/* do it only once! (dofs are visited more than once) */	\
	ldof[j] = - k - 1;						\
      } }

  void Element::newDOFFct1(const DOFAdmin* admin)
  {
    int     i, j, k, n0, nd, nd0;
    DegreeOfFreedom     *ldof;
  
    int vertices = mesh->getGeo(VERTEX);
    int edges = mesh->getGeo(EDGE); 
    int faces = mesh->getGeo(FACE);

    if ((nd = admin->getNumberOfDOFs(VERTEX)))  {
      nd0 = admin->getNumberOfPreDOFs(VERTEX);
      n0 = admin->getMesh()->getNode(VERTEX);
      for (i = 0; i < vertices; i++) {
	CHANGE_DOFS_1(this);
      }
    }

    if(mesh->getDim() > 1) {
      if ((nd = admin->getNumberOfDOFs(EDGE)))  {
	nd0 = admin->getNumberOfPreDOFs(EDGE);
	n0 = admin->getMesh()->getNode(EDGE);
	for (i = 0; i < edges; i++) {
	  CHANGE_DOFS_1(this);
	}
      }
    }

    if (3==mesh->getDim()) {
      if ((nd = admin->getNumberOfDOFs(FACE)))  {
	nd0 = admin->getNumberOfPreDOFs(FACE);
	n0 = admin->getMesh()->getNode(FACE);
	for (i = 0; i < faces; i++) {
	  CHANGE_DOFS_1(this);
	}
      }
    }

    if ((nd = admin->getNumberOfDOFs(CENTER)))  {
      nd0 = admin->getNumberOfPreDOFs(CENTER);
      n0 = admin->getMesh()->getNode(CENTER);
      i = 0;          /* only one center */
      CHANGE_DOFS_1(this);
    }
  }


  void Element::newDOFFct2(const DOFAdmin* admin)
  {
    int     i, j, k, n0, nd, nd0;
    DegreeOfFreedom     *ldof;
   
    int vertices = mesh->getGeo(VERTEX);
    int edges = mesh->getGeo(EDGE); 
    int faces = mesh->getGeo(FACE);

    if ((nd = admin->getNumberOfDOFs(VERTEX)))  {
      nd0 = admin->getNumberOfPreDOFs(VERTEX);
      n0 = admin->getMesh()->getNode(VERTEX);
      for (i = 0; i < vertices; i++) {
	CHANGE_DOFS_2(this);
      }
    }

    if(mesh->getDim() > 1) {
      if ((nd = admin->getNumberOfDOFs(EDGE)))  {
	nd0 = admin->getNumberOfPreDOFs(EDGE);
	n0 = admin->getMesh()->getNode(EDGE);
	for (i = 0; i < edges; i++) {
	  CHANGE_DOFS_2(this);
	}
      }
    }

    if (3==mesh->getDim()) {
      if ((nd = admin->getNumberOfDOFs(FACE)))  {
	nd0 = admin->getNumberOfPreDOFs(FACE);
	n0 = admin->getMesh()->getNode(FACE);
	for (i = 0; i < faces; i++) {
	  CHANGE_DOFS_2(this);
	}
      }
    }

    if ((nd = admin->getNumberOfDOFs(CENTER)))  {
      nd0 = admin->getNumberOfPreDOFs(CENTER);
      n0 = admin->getMesh()->getNode(CENTER);
      i = 0;          /* only one center */
      CHANGE_DOFS_2(this);
    }
  }

#undef CHANGE_DOFS_1
#undef CHANGE_DOFS_2

  /****************************************************************************/
  /* opp_vertex checks whether the face with vertices dof[0],..,dof[DIM-1] is */
  /* part of mel's boundary. returns the opposite vertex if true, -1 else     */
  /****************************************************************************/

  int Element::oppVertex(FixVec<DegreeOfFreedom*, DIMEN> pdof) const
  {
    int      i, j, nv = 0, ov = 0;

    int vertices = mesh->getGeo(VERTEX);
    int dim = mesh->getDim();

    for (i = 0; i < vertices; i++)
      {
	if (nv < i-1)  return(-1);

	for (j = 0; j < dim; j++)
	  {
	    if (dof[i] == pdof[j])
	      {
		/****************************************************************************/
		/* i is a common vertex                                                     */
		/****************************************************************************/
		ov += i;
		nv++;
		break;
	      }
	  }

      }
    if (nv != mesh->getDim()) return(-1);
    /****************************************************************************/
    /*  the opposite vertex is 3(6) - (sum of indices of common vertices) in    */
    /*  2d(3d)                                                                  */
    /****************************************************************************/

    switch(mesh->getDim()) {
    case 1:
      return ov;
      break;
    case 2:
      return 3-ov;
      break;
    case 3:
      return 6-ov;
      break;
    default:
      ERROR_EXIT("invalid dim\n");
      return 0;
    }
  }

  double Element::getNewCoord(int j) const {
    if (j>=0) {
      TEST_EXIT(newCoord)("newCoord = NULL\n");
      return (*newCoord)[j];
    } else { 
      return (newCoord!=NULL);
    }
  }

  void Element::eraseNewCoord() {
    if (newCoord!=NULL) {
      DELETE newCoord;
      newCoord=NULL;
    };
  }

  void Element::serialize(::std::ostream &out) 
  {
    // write children
    if(child[0]) {
      out << child[0]->getTypeName() << ::std::endl;
      child[0]->serialize(out);
      child[1]->serialize(out);
    } else {
      out << "NULL" << ::std::endl;
    }

    // write dofs
    int dim = mesh->getDim();
    int nodes = mesh->getNumberOfNodes();
    out.write(reinterpret_cast<const char*>(&nodes), sizeof(int));

    int i, j = 0, ndof, pos;
    
    for(pos = 0; pos <= dim; pos++) {
      GeoIndex position = INDEX_OF_DIM(pos ,dim);

      ndof = 0;

      for(i = 0; i < mesh->getNumberOfDOFAdmin(); i++) {
	const DOFAdmin *localAdmin = &mesh->getDOFAdmin(i);
	TEST_EXIT(localAdmin)("no admin[%d]\n", i);
	ndof += localAdmin->getNumberOfDOFs(position);	
      }

      if(ndof > 0) {
	for(i = 0; i < mesh->getGeo(position); i++) {
	  if(dof[j] != NULL) {
	    if(Mesh::serializedDOFs[dof[j][0]] == NULL) {
	      Mesh::serializedDOFs[dof[j][0]] = dof[j];
	      out.write(reinterpret_cast<const char*>(&ndof), sizeof(int));
	      out.write(reinterpret_cast<const char*>(dof[j]), 
			ndof * sizeof(DegreeOfFreedom));
	    } else {
	      int minusOne = -1;
	      out.write(reinterpret_cast<const char*>(&minusOne), sizeof(int));
	      out.write(reinterpret_cast<const char*>(&(dof[j][0])), sizeof(DegreeOfFreedom));
	    }
	  } else {
	    int zero = 0;
	    out.write(reinterpret_cast<const char*>(&zero), sizeof(int));
	  }
	  j++;
	}
      }
    }

    // write index
    out.write(reinterpret_cast<const char*>(&index), sizeof(int));

    // write mark
    out.write(reinterpret_cast<const char*>(&mark), sizeof(signed char));

    // write newCoord
    if(newCoord) {
      out << "WorldVector" << ::std::endl;
      newCoord->serialize(out);
    } else {
      out << "NULL" << ::std::endl;
    }

    // write element data
    if(elementData) {
      out << elementData->getTypeName() << ::std::endl;
      elementData->serialize(out);
    } else {
      out << "NULL" << ::std::endl;    
    }
  }

  void Element::deserialize(::std::istream &in)
  {
    ::std::string typeName;

    // read children
    in >> typeName;
    in.get();

    if(typeName != "NULL") {
      if(typeName == "Line") {
	child[0] = new Line(NULL);
	child[1] = new Line(NULL);      
      };
      if(typeName == "Triangle") {
	child[0] = new Triangle(NULL);
	child[1] = new Triangle(NULL);      
      };
      if(typeName == "Tetrahedron") {
	child[0] = new Tetrahedron(NULL);
	child[1] = new Tetrahedron(NULL);      
      };
      child[0]->deserialize(in);
      child[1]->deserialize(in);
    } else {
      child[0] = child[1] = NULL;
    }

    // read dofs
    int nodes;
    in.read(reinterpret_cast<char*>(&nodes), sizeof(int));

    dof = GET_MEMORY(DegreeOfFreedom*, nodes); 

    int i;
    for(i = 0; i < nodes; i++) {
      int dofs;
      in.read(reinterpret_cast<char*>(&dofs), sizeof(int));

      if(dofs) {
	if(dofs != -1) {
	  dof[i] = GET_MEMORY(DegreeOfFreedom, dofs);
	  in.read(reinterpret_cast<char*>(dof[i]), dofs * sizeof(DegreeOfFreedom));
	  if(Mesh::serializedDOFs[dof[i][0]] != NULL) {
	    //WARNING("christina -> dofs already deserialized\n");
	    DegreeOfFreedom *dofPtr = Mesh::serializedDOFs[dof[i][0]];
	    FREE_MEMORY(dof[i], DegreeOfFreedom, dofs);
	    dof[i] = dofPtr;
	  }
	  Mesh::serializedDOFs[dof[i][0]] = dof[i];
	} else {
	  DegreeOfFreedom index;
	  in.read(reinterpret_cast<char*>(&index), sizeof(DegreeOfFreedom));
	  dof[i] = Mesh::serializedDOFs[index];
	}
      } else {
	dof[i] = NULL;
      }
    }
    
    // read index
    in.read(reinterpret_cast<char*>(&index), sizeof(int));

    // write mark
    in.read(reinterpret_cast<char*>(&mark), sizeof(signed char));

    // read newCoord
    in >> typeName;
    in.get();

    if(typeName != "NULL") {
      if(typeName == "WorldVector") {
	newCoord = NEW WorldVector<double>;
	newCoord->deserialize(in);
      } else {
	ERROR_EXIT("unexpected type name\n");
      }
    } else {
      newCoord = NULL;
    }

    // read element data
    in >> typeName;
    in.get();

    if(typeName != "NULL") {
      elementData = CreatorMap<ElementData>::getCreator(typeName)->create();
      if(elementData) {
	elementData->deserialize(in);
      } else {
	ERROR_EXIT("unexpected type name\n");
      }
    } else {
      elementData = NULL;
    }
  }

}
