// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file GMResSolver.h */

#ifndef AMDIS_GMRESSOLVER_H
#define AMDIS_GMRESSOLVER_H

#include "OEMSolver.h"
#include "MemoryManager.h"

namespace AMDiS {

  // ============================================================================
  // ===== class GMResSolver ====================================================
  // ============================================================================

  /** 
   * \ingroup Solver
   *
   * \brief
   * Solves a linear system by the GMRes method with restart and can be used for
   * regular system matrices. 
   */
  template<typename VectorType>
  class GMResSolver : public OEMSolver<VectorType>
  {
  public:
    MEMORY_MANAGED(GMResSolver<VectorType>);

    /** \brief
     * Creator class used in the OEMSolverMap.
     */
    class Creator : public OEMSolverCreator<VectorType>
    {
    public:
      MEMORY_MANAGED(Creator);

      virtual ~Creator() {};

      /** \brief
       * Returns a new GMResSolver object.
       */   
      OEMSolver<VectorType>* create() { 
	return NEW GMResSolver<VectorType>(this->name);
      };
    };

    /** \brief
     * constructor
     */
    GMResSolver(::std::string name);

    /** \brief
     * destructor
     */
    ~GMResSolver();

  protected:
    /** \brief
     * realisation of OEMSolver::solveSystem
     */
    int solveSystem(MatVecMultiplier<VectorType> *mv, VectorType *x, VectorType *b);

    /** \brief
     * realisation of OEMSolver::init
     */
    void init();

    /** \brief
     * realisation of OEMSolver::exit
     */
    void exit();

  private:
    /** \brief
     * internal used method
     */
    double householderVector(double sigma, 
			     VectorType *u, 
			     VectorType *x, 
			     int offset);

    /** \brief
     * internal used method
     */
    void newBasisVector(int m, int k, VectorType **U, double **LR, 
			MatVecMultiplier<VectorType> *matVec,
			VectorType *v, VectorType *b, double* y);

    /** \brief
     * internal used method
     */
    int solve_k(MatVecMultiplier<VectorType> *mv,
		VectorType *x, const VectorType *b);


  private:
    double TOL;
    int restart;
    int dim;

    // pointer to memory needed for solveSystem
    VectorType *r, *v;
    VectorType **U;
    double **LR;
    double **givens, *w, *y;

    double gmres_k_residual_0;
  };


}

#include "GMResSolver.hh"

#endif // AMDIS_GMRESSOLVER_H
