// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file Lagrange.h */

#ifndef AMDIS_LAGRANGE_H
#define AMDIS_LAGRANGE_H

#include "BasisFunction.h"
#include "FixVec.h"
#include "AbstractFunction.h"
#include "MemoryManager.h"
#include <list>

namespace AMDiS {

#define MAX_DIM 3
#define MAX_DEGREE 4

  template<typename ReturnType, typename ArgumentType> class AbstractFunction;

  // ============================================================================
  // ===== class Lagrange =======================================================
  // ============================================================================

  /** \ingroup FEMSpace
   * \brief
   * Lagrange basis functions. Sub class of BasisFunction
   */
  class Lagrange : public BasisFunction
  {
  public:
    MEMORY_MANAGED(Lagrange);

  protected:
    /** \brief
     * Constructs lagrange basis functions with the given dim and degree.
     * Constructor is protected to avoid multiple instantiation of identical
     * basis functions. Use \ref getLagrange instead.
     */
    Lagrange(int dim_, int degree_);

    /** \brief
     * destructor
     */
    virtual ~Lagrange();

  public:
    /** \brief
     * Returns a pointer to lagrange basis functions with the given dim and
     * degree. Multiple instantiation of identical basis functions is avoided
     * by rembering once created basis functions in \ref allBasFcts.
     */
    static Lagrange* getLagrange(int dim_, int degree_);

    /** \brief
     * Implements BasisFunction::interpol
     */
    const double   *interpol(const ElInfo *, int, const int *, 
			     AbstractFunction<double, WorldVector<double> >*, double *);


    /** \brief
     * Implements BasisFunction::interpol
     */
    const WorldVector<double> *interpol(const ElInfo *, int, const int *b_no,
					AbstractFunction<WorldVector<double>, WorldVector<double> >*, 
					WorldVector<double> *);

    /** \brief
     * Returns the barycentric coordinates of the i-th basis function.
     */
    DimVec<double> *getCoords(int i) const;

    /** \brief
     * Implements BasisFunction::getDOFIndices
     */
    const DegreeOfFreedom* getDOFIndices(const Element*, const DOFAdmin&, 
					 DegreeOfFreedom*) const;

    /** \brief
     * Implements BasisFunction::getBound
     */
    const BoundaryType* getBound(const ElInfo*, BoundaryType *) const;

    /** \brief
     * Calculates the local vertex indices which are involved in evaluating
     * the nodeIndex-th DOF at the positionIndex-th part of type position 
     * (VERTEX/EDGE/FACE/CENTER). nodeIndex determines the permutation
     * of the involved vertices. So in 1d for lagrange4 there are two DOFs at
     * the CENTER (which is an edge in this case). Then vertices[0] = {0, 1} and
     * vertices[1] = {1, 0}. This allows to use the same local basis function 
     * for all DOFs at the same position.
     */
    static void setVertices(int dim, int degree, 
			    GeoIndex position, int positionIndex, int nodeIndex, 
			    int** vertices);

    /** \brief
     * Implements BasisFunction::refineInter
     */
    inline void  refineInter(DOFIndexed<double> *drv,
			     RCNeighbourList* list, 
			     int n) 
    {
      if(refineInter_fct)
	(*refineInter_fct)(drv, list, n, this);
    };

    /** \brief
     * Implements BasisFunction::coarseRestrict
     */
    inline void  coarseRestr(DOFIndexed<double> *drv,
			     RCNeighbourList* list,
			     int n)
    {
      if(coarseRestr_fct)
	(*coarseRestr_fct)(drv, list, n, this);
    };
  
    /** \brief
     * Implements BasisFunction::coarseInter
     */
    inline void  coarseInter(DOFIndexed<double> *drv, 
			     RCNeighbourList* list, 
			     int n) 
    {
      if(coarseInter_fct)
	(*coarseInter_fct)(drv, list, n, this);
    };
  
    /** \brief
     * Implements BasisFunction::getLocalIndices().
     */
    const DegreeOfFreedom *getLocalIndices(const Element*,
					   const DOFAdmin*,
					   DegreeOfFreedom*) const;

    /** \brief
     * Implements BasisFunction::getVec().
     */
    //   const double *getVec(const Element* el, 
    // 		       const DOFVector<double> * dv, 
    // 		       double * d) const;

    /** \brief
     * Implements BasisFunction::getVec
     */
    //   const WorldVector<double> *getVec(const Element* el, 
    // 				    const DOFVector<WorldVector<double> > * dv, 
    // 				    WorldVector<double> *d) const;

    /** \brief
     * Implements BasisFunction::l2ScpFctBas
     */
    void l2ScpFctBas(Quadrature* q,
		     AbstractFunction<double, WorldVector<double> >* f,
		     DOFVector<double>* fh);

    /** \brief
     * Implements BasisFunction::l2ScpFctBas
     */
    void l2ScpFctBas(Quadrature* q,
		     AbstractFunction<WorldVector<double>, WorldVector<double> >* f,
		     DOFVector<WorldVector<double> >* fh);

  protected:
    /** \brief
     * sets the barycentric coordinates (stored in \ref bary) of the local 
     * basis functions.
     */
    void setBary();

    /** \brief
     * Recursive calculation of coordinates. Used by \ref setBary
     */
    void createCoords(int* coordInd, int numCoords, int dimIndex, int rest, 
		      DimVec<double>* vec=NULL);

    /** \brief
     * Used by \ref setBary
     */
    int** getIndexPermutations(int numIndices) const;

    /** \brief
     * Implements BasisFunction::setNDOF
     */
    void setNDOF();

    /** \brief
     * Sets used function pointers
     */
    void setFunctionPointer();

    /** \brief
     * Used by \ref getDOFIndices and \ref getVec
     */
    int* orderOfPositionIndices(const Element* el, GeoIndex position, 
				int positionIndex) const;

    /** \brief
     * Calculates the number of DOFs needed for Lagrange of the given dim and
     * degree.
     */
    static int getNumberOfDOFs(int dim, int degree);

  private:
    /** \brief
     * barycentric coordinates of the locations of all basis functions
     */
    ::std::vector<DimVec<double>* > *bary;

    /** \name static dim-degree-arrays
     * \{
     */
    static ::std::vector<DimVec<double>* > baryDimDegree[MAX_DIM+1][MAX_DEGREE+1];
    static DimVec<int>* ndofDimDegree[MAX_DIM+1][MAX_DEGREE+1];
    static int nBasFctsDimDegree[MAX_DIM+1][MAX_DEGREE+1];
    static ::std::vector<BasFctType*> phiDimDegree[MAX_DIM+1][MAX_DEGREE+1];
    static ::std::vector<GrdBasFctType*> grdPhiDimDegree[MAX_DIM+1][MAX_DEGREE+1];
    static ::std::vector<D2BasFctType*> D2PhiDimDegree[MAX_DIM+1][MAX_DEGREE+1];
    /** \} */

    /** \brief
     * List of all used BasisFunctions in the whole program. Avoids duplicate
     * instantiation of identical BasisFunctions. 
     */
    static ::std::list<Lagrange*> allBasFcts;


  protected:
    /** \brief
     * Pointer to the used refineInter function
     */
    void  (*refineInter_fct)(DOFIndexed<double> *, RCNeighbourList*, int, 
			     BasisFunction*);

    /** \name refineInter functions
     * \{
     */  
    static void  refineInter0(DOFIndexed<double> *, RCNeighbourList*, int, 
			      BasisFunction*);
    static void  refineInter1(DOFIndexed<double> *, RCNeighbourList*, int, 
			      BasisFunction*);
    static void  refineInter2_1d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    static void  refineInter2_2d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    static void  refineInter2_3d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    static void  refineInter3_1d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    static void  refineInter3_2d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    static void  refineInter3_3d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    static void  refineInter4_1d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    static void  refineInter4_2d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    static void  refineInter4_3d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    /** \} */

    /** \brief
     * Pointer to the used coarseRestr function
     */
    void  (*coarseRestr_fct)(DOFIndexed<double> *, RCNeighbourList*, int, 
			     BasisFunction*);

    /** \name coarseRestr functions
     * \{
     */  
    static void  coarseRestr0(DOFIndexed<double> *, RCNeighbourList*, int, 
			      BasisFunction*);
    static void  coarseRestr1(DOFIndexed<double> *, RCNeighbourList*, int, 
			      BasisFunction*);
    static void  coarseRestr2_1d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    static void  coarseRestr2_2d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    static void  coarseRestr2_3d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    static void  coarseRestr3_1d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    static void  coarseRestr3_2d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    static void  coarseRestr3_3d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    static void  coarseRestr4_1d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    static void  coarseRestr4_2d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    static void  coarseRestr4_3d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    /** \} */

    /** \brief
     * Pointer to the used coarseInter function
     */
    void  (*coarseInter_fct)(DOFIndexed<double> *, RCNeighbourList*, int, 
			     BasisFunction*);  

    /** \name coarseInter functions
     * \{
     */
    static void  coarseInter0(DOFIndexed<double> *, RCNeighbourList*, int, 
			      BasisFunction*);
    static void  coarseInter1(DOFIndexed<double> *, RCNeighbourList*, int, 
			      BasisFunction*);
    static void  coarseInter2_1d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    static void  coarseInter2_2d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    static void  coarseInter2_3d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    static void  coarseInter3_1d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    static void  coarseInter3_2d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    static void  coarseInter3_3d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    static void  coarseInter4_1d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    static void  coarseInter4_2d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    static void  coarseInter4_3d(DOFIndexed<double> *, RCNeighbourList*, int, 
				 BasisFunction*);
    /** \} */

    // ===== Phi =========================================================

    /** \brief
     * AbstractFunction which implements lagrange basis functions
     */
    class Phi : public BasFctType
    {
    public:
      MEMORY_MANAGED(Phi);

      /** \brief
       * Constructs the local lagrange basis function for the given position,
       * positionIndex and nodeIndex. owner_ is a pointer to the Lagrange object
       * this basis function belongs to.
       */
      Phi(Lagrange* owner_, GeoIndex position, int positionIndex, int nodeIndex);

      /** \brief
       * Destructor
       */
      virtual ~Phi();

    private:
      /** \brief
       * Lagrange object this function belongs to
       */
      Lagrange* owner;

      /** \brief
       * vertices needed for evaluation of this function
       */
      int* vertices;
    
      /** \brief
       * Pointer to the evaluating function
       */
      const double& (*func)(const DimVec<double>& lambda, int* vert);

      /** \brief
       * Returns \ref func(lambda, vertices)
       */
      inline const double& operator()(const DimVec<double>& lambda) const {
	return func(lambda, vertices);
      };

      /** \name basis functions for different degrees
       * \{
       */

      // ====== Lagrange0 ================================================
      // center
      inline static const double& phi0c(const DimVec<double>& /*lambda*/,
					int* /*vertices*/) {
	static double phi = 1.0;
	return phi;
      };

      // ====== Lagrange1 ================================================
      // vertex
      inline static const double& phi1v(const DimVec<double>& lambda, 
					int* vertices) {
	static double phi = 0;
	phi = lambda[vertices[0]]; 
	return phi; 
      };

      // ====== Lagrange2 ================================================
      // vertex
      inline static const double& phi2v(const DimVec<double>& lambda, 
					int* vertices) {
	static double phi = 0;
	phi = lambda[vertices[0]]*(2.0*lambda[vertices[0]] - 1.0);
	return phi; 
      };    

      // edge
      inline static const double& phi2e(const DimVec<double>& lambda, 
					int* vertices) {
	static double phi = 0;
	phi = (4.0*lambda[vertices[0]]*lambda[vertices[1]]);
	return phi; 
      };

      // ====== Lagrange3 ================================================
      // vertex
      inline static const double& phi3v(const DimVec<double>& lambda, 
					int* vertices) {
	static double phi = 0;
	phi = (4.5*(lambda[vertices[0]] - 1.0)*lambda[vertices[0]] + 1.0) * 
	  lambda[vertices[0]];
	return phi; 
      };

      // edge
      inline static const double& phi3e(const DimVec<double>& lambda, 
					int* vertices) {
	static double phi = 0;
	phi = (13.5*lambda[vertices[0]] - 4.5) * 
	  lambda[vertices[0]]*lambda[vertices[1]];
	return phi; 
      };

      // face
      inline static const double& phi3f(const DimVec<double>& lambda, 
					int* vertices) {
	static double phi = 0;
	phi = 27.0*lambda[vertices[0]]*lambda[vertices[1]]*lambda[vertices[2]];
	return phi; 
      };

      // ====== Lagrange4 ================================================
      // vertex
      inline static const double& phi4v(const DimVec<double>& lambda, 
					int* vertices) {
	static double phi = 0;
	phi = (((32.0*lambda[vertices[0]] - 48.0)*lambda[vertices[0]] + 22.0)
	       *lambda[vertices[0]] - 3.0)*lambda[vertices[0]]/3.0;
	return phi; 
      };

      // edge
      inline static const double& phi4e0(const DimVec<double>& lambda, 
					 int* vertices) {
	static double phi = 0;
	phi = ((128.0*lambda[vertices[0]] - 96.0) * lambda[vertices[0]] + 16.0)
	  * lambda[vertices[0]]*lambda[vertices[1]]/3.0;
	return phi; 
      };

      inline static const double& phi4e1(const DimVec<double>& lambda, 
					 int* vertices) {
	static double phi = 0;
	phi = (4.0*lambda[vertices[0]] - 1.0)*lambda[vertices[0]] * 
	  (4.0*lambda[vertices[1]] - 1.0)*lambda[vertices[1]]*4.0;
	return phi; 
      };

      // face
      inline static const double& phi4f(const DimVec<double>& lambda, 
					int* vertices) {
	static double phi = 0;
	phi = (4.0*lambda[vertices[0]] - 1.0)*lambda[vertices[0]] * 
	  lambda[vertices[1]]*lambda[vertices[2]]*32.0;
	return phi; 
      };

      // center
      inline static const double& phi4c(const DimVec<double>& lambda, 
					int* vertices) {
	static double phi = 0;
	phi = 256.0*lambda[vertices[0]]*lambda[vertices[1]] * 
	  lambda[vertices[2]]*lambda[vertices[3]];
	return phi; 
      };

    };
  
    /** \} */

    // ===== GrdPhi ======================================================

    /** \brief
     * AbstractFunction which implements gradients of lagrange basis functions.
     * See \ref Phi
     */
    class GrdPhi : public GrdBasFctType
    {
    public:
      MEMORY_MANAGED(GrdPhi);

      GrdPhi(Lagrange* owner_, GeoIndex position, 
	     int positionIndex, 
	     int nodeIndex);
      virtual ~GrdPhi();
    private:
      Lagrange* owner;
      int* vertices;
      const DimVec<double>& (*func)(const DimVec<double>& lambda,
				    int* vertices_,
				    Lagrange* owner_);

      inline const DimVec<double>& operator()(const DimVec<double>& lambda) const {
	return func(lambda, vertices, owner);
      };

      // ====== Lagrange0 ================================================
      // center
      inline static const DimVec<double>& grdPhi0c(const DimVec<double>& /* lambda*/, 
						   int* /*vertices*/, 
						   Lagrange* owner) {
	static DimVec<double>* grd[4]={NULL,
				       NEW DimVec<double>(1, DEFAULT_VALUE, 0.),
				       NEW DimVec<double>(2, DEFAULT_VALUE, 0.),
				       NEW DimVec<double>(3, DEFAULT_VALUE, 0.)};
	return(*(grd[owner->getDim()]));
      };

      // ====== Lagrange1 ================================================
      // vertex
      inline static const DimVec<double>& grdPhi1v(const DimVec<double>& /*lambda*/, 
						   int* vertices, 
						   Lagrange* owner) {
	static DimVec<double>* grd[4]={NULL,
				       NEW DimVec<double>(1, DEFAULT_VALUE, 0.),
				       NEW DimVec<double>(2, DEFAULT_VALUE, 0.),
				       NEW DimVec<double>(3, DEFAULT_VALUE, 0.)};
	int dim = owner->getDim();
	grd[dim]->set(0.);
	(*grd[dim])[vertices[0]] = 1.0;
	return(*(grd[dim]));
      };

      // ====== Lagrange2 ================================================
      // vertex
      inline static const DimVec<double>& grdPhi2v(const DimVec<double>& lambda, 
						   int* vertices, 
						   Lagrange* owner) {
	static DimVec<double>* grd[4]={NULL,
				       NEW DimVec<double>(1, DEFAULT_VALUE, 0.),
				       NEW DimVec<double>(2, DEFAULT_VALUE, 0.),
				       NEW DimVec<double>(3, DEFAULT_VALUE, 0.)};
	int dim = owner->getDim();
	grd[dim]->set(0.);
	(*grd[dim])[vertices[0]] = 4.0*lambda[vertices[0]] - 1.0;
	return(*(grd[dim]));
      };

      // edge
      inline static const DimVec<double>& grdPhi2e(const DimVec<double>& lambda, 
						   int* vertices, 
						   Lagrange* owner) {
	static DimVec<double>* grd[4]={NULL,
				       NEW DimVec<double>(1, DEFAULT_VALUE, 0.),
				       NEW DimVec<double>(2, DEFAULT_VALUE, 0.),
				       NEW DimVec<double>(3, DEFAULT_VALUE, 0.)};
	int dim = owner->getDim();
	grd[dim]->set(0.);
	(*grd[dim])[vertices[0]] = 4.0*lambda[vertices[1]];
	(*grd[dim])[vertices[1]] = 4.0*lambda[vertices[0]];
	return(*(grd[dim]));
      };

      // ===== Lagrange3 ================================================
      // vertex
      inline static const DimVec<double>& grdPhi3v(const DimVec<double>& lambda,
						   int* vertices,
						   Lagrange* owner) {
	static DimVec<double>* grd[4]={NULL,
				       NEW DimVec<double>(1, DEFAULT_VALUE, 0.),
				       NEW DimVec<double>(2, DEFAULT_VALUE, 0.),
				       NEW DimVec<double>(3, DEFAULT_VALUE, 0.)};
	int dim = owner->getDim();
	grd[dim]->set(0.);
	(*grd[dim])[vertices[0]] = (13.5*lambda[vertices[0]] - 9.0) * 
	  lambda[vertices[0]] + 1;
	return(*(grd[dim]));
      };

      // edge
      inline static const DimVec<double>& grdPhi3e(const DimVec<double>& lambda,
						   int* vertices, 
						   Lagrange* owner) {
	static DimVec<double>* grd[4]={NULL,
				       NEW DimVec<double>(1, DEFAULT_VALUE, 0.),
				       NEW DimVec<double>(2, DEFAULT_VALUE, 0.),
				       NEW DimVec<double>(3, DEFAULT_VALUE, 0.)};
	int dim = owner->getDim();
	grd[dim]->set(0.);
	(*grd[dim])[vertices[0]] = (27.0*lambda[vertices[0]] - 4.5) * 
	  lambda[vertices[1]];
	(*grd[dim])[vertices[1]] = (13.5*lambda[vertices[0]] - 4.5) * 
	  lambda[vertices[0]];
	return(*(grd[dim]));
      };

      // face
      inline static const DimVec<double>& grdPhi3f(const DimVec<double>& lambda, 
						   int* vertices, 
						   Lagrange* owner) {
	static DimVec<double>* grd[4]={NULL,
				       NEW DimVec<double>(1, DEFAULT_VALUE, 0.),
				       NEW DimVec<double>(2, DEFAULT_VALUE, 0.),
				       NEW DimVec<double>(3, DEFAULT_VALUE, 0.)};
	int dim = owner->getDim();
	grd[dim]->set(0.);
	(*grd[dim])[vertices[0]] = 27.0*lambda[vertices[1]]*lambda[vertices[2]];
	(*grd[dim])[vertices[1]] = 27.0*lambda[vertices[0]]*lambda[vertices[2]];
	(*grd[dim])[vertices[2]] = 27.0*lambda[vertices[0]]*lambda[vertices[1]];
	return(*(grd[dim]));
      };

    
      // ===== Lagrange4 ================================================
      // vertex
      inline static const DimVec<double>& grdPhi4v(const DimVec<double>& lambda, 
						   int* vertices, 
						   Lagrange* owner) {
	static DimVec<double>* grd[4]={NULL,
				       NEW DimVec<double>(1, DEFAULT_VALUE, 0.),
				       NEW DimVec<double>(2, DEFAULT_VALUE, 0.),
				       NEW DimVec<double>(3, DEFAULT_VALUE, 0.)};
	int dim = owner->getDim();
	grd[dim]->set(0.);
	(*grd[dim])[vertices[0]] = 
	  ((128.0*lambda[vertices[0]] - 144.0) * lambda[vertices[0]] + 44.0) * 
	  lambda[vertices[0]]/3.0 - 1.0;
	return(*(grd[dim]));
      };
    
      // edge
      inline static const DimVec<double>& grdPhi4e0(const DimVec<double>& lambda,
						    int* vertices, 
						    Lagrange* owner) {
	static DimVec<double>* grd[4]={NULL,
				       NEW DimVec<double>(1, DEFAULT_VALUE, 0.),
				       NEW DimVec<double>(2, DEFAULT_VALUE, 0.),
				       NEW DimVec<double>(3, DEFAULT_VALUE, 0.)};
	int dim = owner->getDim();
	grd[dim]->set(0.);
	(*grd[dim])[vertices[0]] = ((128*lambda[vertices[0]] - 64.0) * 
				    lambda[vertices[0]] + 16.0/3.0)*lambda[vertices[1]];
	(*grd[dim])[vertices[1]] = ((128*lambda[vertices[0]] - 96.0) * 
				    lambda[vertices[0]] + 16.0)*lambda[vertices[0]]/3.0;
	return(*(grd[dim]));
      };

      inline static const DimVec<double>& grdPhi4e1(const DimVec<double>& lambda,
						    int* vertices,
						    Lagrange* owner) {
	static DimVec<double>* grd[4]={NULL,
				       NEW DimVec<double>(1, DEFAULT_VALUE, 0.),
				       NEW DimVec<double>(2, DEFAULT_VALUE, 0.),
				       NEW DimVec<double>(3, DEFAULT_VALUE, 0.)};
	int dim = owner->getDim();
	grd[dim]->set(0.);
	(*grd[dim])[vertices[0]] = 4.0*(8.0*lambda[vertices[0]] - 1.0) * 
	  lambda[vertices[1]]*(4.0*lambda[vertices[1]] - 1.0);
	(*grd[dim])[vertices[1]] = 4.0*lambda[vertices[0]] * 
	  (4.0*lambda[vertices[0]] - 1.0)*(8.0*lambda[vertices[1]] - 1.0);
	return(*(grd[dim]));
      };

      // face
      inline static const DimVec<double>& grdPhi4f(const DimVec<double>& lambda, 
						   int* vertices, 
						   Lagrange* owner) {
	static DimVec<double>* grd[4]={NULL,
				       NEW DimVec<double>(1, DEFAULT_VALUE, 0.),
				       NEW DimVec<double>(2, DEFAULT_VALUE, 0.),
				       NEW DimVec<double>(3, DEFAULT_VALUE, 0.)};
	int dim = owner->getDim();
	grd[dim]->set(0.);
	(*grd[dim])[vertices[0]] = 32.0*(8.0*lambda[vertices[0]] - 1.0) * 
	  lambda[vertices[1]]*lambda[vertices[2]];
	(*grd[dim])[vertices[1]] = 32.0*(4.0*lambda[vertices[0]] - 1.0) * 
	  lambda[vertices[0]]*lambda[vertices[2]];
	(*grd[dim])[vertices[2]] = 32.0*(4.0*lambda[vertices[0]] - 1.0) * 
	  lambda[vertices[0]]*lambda[vertices[1]];
	return(*(grd[dim]));
      };
    
      // center
      inline static const DimVec<double>& grdPhi4c(const DimVec<double>& lambda, 
						   int* vertices,
						   Lagrange* owner) {
	static DimVec<double>* grd[4]={NULL,
				       NEW DimVec<double>(1, DEFAULT_VALUE, 0.),
				       NEW DimVec<double>(2, DEFAULT_VALUE, 0.),
				       NEW DimVec<double>(3, DEFAULT_VALUE, 0.)};
	int dim = owner->getDim();
	grd[dim]->set(0.);
	(*grd[dim])[0] = 256.0*lambda[vertices[1]]*lambda[vertices[2]] * 
	  lambda[vertices[3]];
	(*grd[dim])[1] = 256.0*lambda[vertices[0]]*lambda[vertices[2]] * 
	  lambda[vertices[3]];
	(*grd[dim])[2] = 256.0*lambda[vertices[0]]*lambda[vertices[1]] * 
	  lambda[vertices[3]];
	(*grd[dim])[3] = 256.0*lambda[vertices[0]]*lambda[vertices[1]] * 
	  lambda[vertices[2]];
	return(*(grd[dim]));
      };
    };

    
    // ===== D2Phi =======================================================

    /** \brief
     * AbstractFunction which implements second derivatives of Lagrange basis
     * functions. See \ref Phi
     */
    class D2Phi : public D2BasFctType
    {
    public:
      MEMORY_MANAGED(D2Phi);

      D2Phi(Lagrange* owner_, GeoIndex position, 
	    int positionIndex, int nodeIndex);

      virtual ~D2Phi();
    private:
      Lagrange* owner;
      int* vertices;
    
      const DimMat<double>& (*func)(const DimVec<double>& lambda,
				    int* vertices_,
				    Lagrange* owner_);

      inline const DimMat<double>& operator()(const DimVec<double>& lambda) const {
	return func(lambda, vertices, owner);
      };

      // ===== Lagrange0 ================================================
      // center
      inline static const DimMat<double>& D2Phi0c(const DimVec<double>& /*lambda*/, 
						  int* /*vertices*/, 
						  Lagrange* owner) {
	static DimMat<double>* D2[4]={NULL,
				      NEW DimMat<double>(1, DEFAULT_VALUE, 0.0),
				      NEW DimMat<double>(2, DEFAULT_VALUE, 0.0),
				      NEW DimMat<double>(3, DEFAULT_VALUE, 0.0)};
	int dim = owner->getDim();
	return (*D2[dim]);
      };

      // ===== Lagrange1 ================================================
      // vertex
      inline static const DimMat<double>& D2Phi1v(const DimVec<double>& /*lambda*/, 
						  int* /*vertices*/, 
						  Lagrange* owner) {
	static DimMat<double>* D2[4]={NULL,
				      NEW DimMat<double>(1, DEFAULT_VALUE, 0.0),
				      NEW DimMat<double>(2, DEFAULT_VALUE, 0.0),
				      NEW DimMat<double>(3, DEFAULT_VALUE, 0.0)};
	int dim = owner->getDim();
	return (*D2[dim]);
      };

      // ===== Lagrange2 ================================================
      // vertex
      inline static const DimMat<double>& D2Phi2v(const DimVec<double>& /*lambda*/, 
						  int* vertices, 
						  Lagrange* owner) {
	static DimMat<double>* D2[4]={NULL,
				      NEW DimMat<double>(1, DEFAULT_VALUE, 0.0),
				      NEW DimMat<double>(2, DEFAULT_VALUE, 0.0),
				      NEW DimMat<double>(3, DEFAULT_VALUE, 0.0)};
	int dim = owner->getDim();
	D2[dim]->set(0.);
	(*D2[dim])[vertices[0]][vertices[0]] = 4.0;
	return (*D2[dim]);
      };

      // edge
      inline static const DimMat<double>& D2Phi2e(const DimVec<double>& /*lambda*/, 
						  int* vertices, 
						  Lagrange* owner) {
	static DimMat<double>* D2[4]={NULL,
				      NEW DimMat<double>(1, DEFAULT_VALUE, 0.0),
				      NEW DimMat<double>(2, DEFAULT_VALUE, 0.0),
				      NEW DimMat<double>(3, DEFAULT_VALUE, 0.0)};
	int dim = owner->getDim();
	D2[dim]->set(0.);
	(*D2[dim])[vertices[0]][vertices[1]] = 4.0;
	(*D2[dim])[vertices[1]][vertices[0]] = 4.0;
	return (*D2[dim]);
      };


      // ===== Lagrange3 ================================================
      // vertex
      inline static const DimMat<double>& D2Phi3v(const DimVec<double>& lambda, 
						  int* vertices, 
						  Lagrange* owner) {
	static DimMat<double>* D2[4]={NULL,
				      NEW DimMat<double>(1, DEFAULT_VALUE, 0.0),
				      NEW DimMat<double>(2, DEFAULT_VALUE, 0.0),
				      NEW DimMat<double>(3, DEFAULT_VALUE, 0.0)};
	int dim = owner->getDim();
	D2[dim]->set(0.);
	(*D2[dim])[vertices[0]][vertices[0]] = 27.0*lambda[vertices[0]] - 9.0;
	return (*D2[dim]);
      };

      // edge
      inline static const DimMat<double>& D2Phi3e(const DimVec<double>& lambda, 
						  int* vertices, 
						  Lagrange* owner) {
	static DimMat<double>* D2[4]={NULL,
				      NEW DimMat<double>(1, DEFAULT_VALUE, 0.0),
				      NEW DimMat<double>(2, DEFAULT_VALUE, 0.0),
				      NEW DimMat<double>(3, DEFAULT_VALUE, 0.0)};
	int dim = owner->getDim();
	D2[dim]->set(0.);
	(*D2[dim])[vertices[0]][vertices[0]] = 27.0*lambda[vertices[1]];
	(*D2[dim])[vertices[0]][vertices[1]] = 
	  (*D2[dim])[vertices[1]][vertices[0]] = 27.0*lambda[vertices[0]] - 4.5;
	return (*D2[dim]);
      };

      // face
      inline static const DimMat<double>& D2Phi3f(const DimVec<double>& lambda, 
						  int* vertices, 
						  Lagrange* owner) {
	static DimMat<double>* D2[4]={NULL,
				      NEW DimMat<double>(1, DEFAULT_VALUE, 0.0),
				      NEW DimMat<double>(2, DEFAULT_VALUE, 0.0),
				      NEW DimMat<double>(3, DEFAULT_VALUE, 0.0)};
	int dim = owner->getDim();
	D2[dim]->set(0.);
	(*D2[dim])[vertices[0]][vertices[1]] = 
	  (*D2[dim])[vertices[1]][vertices[0]] = 27.0*lambda[vertices[2]];
	(*D2[dim])[vertices[0]][vertices[2]] = 
	  (*D2[dim])[vertices[2]][vertices[0]] = 27.0*lambda[vertices[1]];
	(*D2[dim])[vertices[1]][vertices[2]] = 
	  (*D2[dim])[vertices[2]][vertices[1]] = 27.0*lambda[vertices[0]];
	return (*D2[dim]);
      };


      // ===== Lagrange4 ================================================
      // vertex
      inline static const DimMat<double>& D2Phi4v(const DimVec<double>& lambda, 
						  int* vertices, 
						  Lagrange* owner) {
	static DimMat<double>* D2[4]={NULL,
				      NEW DimMat<double>(1, DEFAULT_VALUE, 0.0),
				      NEW DimMat<double>(2, DEFAULT_VALUE, 0.0),
				      NEW DimMat<double>(3, DEFAULT_VALUE, 0.0)};
	int dim = owner->getDim();
	D2[dim]->set(0.);
	(*D2[dim])[vertices[0]][vertices[0]] = 
	  (128.0*lambda[vertices[0]] - 96.0)*lambda[vertices[0]] + 44.0/3.0;
	return (*D2[dim]);
      };

      // edge
      inline static const DimMat<double>& D2Phi4e0(const DimVec<double>& lambda, 
						   int* vertices, 
						   Lagrange* owner) {
	static DimMat<double>* D2[4]={NULL,
				      NEW DimMat<double>(1, DEFAULT_VALUE, 0.0),
				      NEW DimMat<double>(2, DEFAULT_VALUE, 0.0),
				      NEW DimMat<double>(3, DEFAULT_VALUE, 0.0)};
	int dim = owner->getDim();
	D2[dim]->set(0.);
	(*D2[dim])[vertices[0]][vertices[0]] = 
	  (256.0*lambda[vertices[0]] - 64.0)*lambda[vertices[1]];
	(*D2[dim])[vertices[0]][vertices[1]] = 
	  (*D2[dim])[vertices[1]][vertices[0]] =
	  (128.0*lambda[vertices[0]] - 64.0)*lambda[vertices[0]] + 16.0/3.0;
	return (*D2[dim]);
      };

      inline static const DimMat<double>& D2Phi4e1(const DimVec<double>& lambda, 
						   int* vertices, 
						   Lagrange* owner) {
	static DimMat<double>* D2[4]={NULL,
				      NEW DimMat<double>(1, DEFAULT_VALUE, 0.0),
				      NEW DimMat<double>(2, DEFAULT_VALUE, 0.0),
				      NEW DimMat<double>(3, DEFAULT_VALUE, 0.0)};
	int dim = owner->getDim();
	D2[dim]->set(0.);
	(*D2[dim])[vertices[0]][vertices[0]] = 
	  32.0*lambda[vertices[1]]*(4.0*lambda[vertices[1]] - 1.0);
	(*D2[dim])[vertices[0]][vertices[1]] = 
	  (*D2[dim])[vertices[1]][vertices[0]] = 
	  4.0*(8.0*lambda[vertices[0]] - 1.0)*(8.0*lambda[vertices[1]] - 1.0);
	(*D2[dim])[vertices[1]][vertices[1]] = 
	  32.0*lambda[vertices[0]]*(4.0*lambda[vertices[0]] - 1.0);
	return (*D2[dim]);
      };

      // face
      inline static const DimMat<double>& D2Phi4f(const DimVec<double>& lambda, 
						  int* vertices, 
						  Lagrange* owner) {
	static DimMat<double>* D2[4]={NULL,
				      NEW DimMat<double>(1, DEFAULT_VALUE, 0.0),
				      NEW DimMat<double>(2, DEFAULT_VALUE, 0.0),
				      NEW DimMat<double>(3, DEFAULT_VALUE, 0.0)};
	int dim = owner->getDim();
	D2[dim]->set(0.);
	(*D2[dim])[vertices[0]][vertices[0]] = 
	  256.0*lambda[vertices[1]]*lambda[vertices[2]];
	(*D2[dim])[vertices[0]][vertices[1]] = 
	  (*D2[dim])[vertices[1]][vertices[0]] = 
	  32.0*(8.0*lambda[vertices[0]] - 1.0)*lambda[vertices[2]];
	(*D2[dim])[vertices[0]][vertices[2]] = 
	  (*D2[dim])[vertices[2]][vertices[0]] = 
	  32.0*(8.0*lambda[vertices[0]] - 1.0)*lambda[vertices[1]];
	(*D2[dim])[vertices[1]][vertices[2]] = 
	  (*D2[dim])[vertices[2]][vertices[1]] = 
	  32.0*(4.0*lambda[vertices[0]] - 1.0)*lambda[vertices[0]];
	return (*D2[dim]);
      };

      // center
      inline static const DimMat<double>& D2Phi4c(const DimVec<double>& lambda, 
						  int* vertices, 
						  Lagrange* owner) {
	static DimMat<double>* D2[4]={NULL,
				      NEW DimMat<double>(1, DEFAULT_VALUE, 0.0),
				      NEW DimMat<double>(2, DEFAULT_VALUE, 0.0),
				      NEW DimMat<double>(3, DEFAULT_VALUE, 0.0)};
	int dim = owner->getDim();
	D2[dim]->set(0.);
	(*D2[dim])[vertices[0]][vertices[1]] = 
	  (*D2[dim])[vertices[1]][vertices[0]] = 
	  256.0*lambda[vertices[2]]*lambda[vertices[3]];
	(*D2[dim])[vertices[0]][vertices[2]] = 
	  (*D2[dim])[vertices[2]][vertices[0]] = 
	  256.0*lambda[vertices[1]]*lambda[vertices[3]];
	(*D2[dim])[vertices[0]][vertices[3]] = 
	  (*D2[dim])[vertices[3]][vertices[0]] = 
	  256.0*lambda[vertices[1]]*lambda[vertices[2]];
	(*D2[dim])[vertices[1]][vertices[2]] = 
	  (*D2[dim])[vertices[2]][vertices[1]] = 
	  256.0*lambda[vertices[0]]*lambda[vertices[3]];
	(*D2[dim])[vertices[1]][vertices[3]] = 
	  (*D2[dim])[vertices[3]][vertices[1]] = 
	  256.0*lambda[vertices[0]]*lambda[vertices[2]];
	(*D2[dim])[vertices[2]][vertices[3]] = 
	  (*D2[dim])[vertices[3]][vertices[2]] = 
	  256.0*lambda[vertices[0]]*lambda[vertices[1]];
	return (*D2[dim]);
      };
    };
  };

}

#endif // AMDIS_LAGRANGE_H
