#include "MacroElement.h"
#include "Boundary.h"
#include "FiniteElemSpace.h"
#include "Mesh.h"
#include <string>
#include "FixVec.h"
#include "FixVecConvert.h"
#include <map>

namespace AMDiS {

  MacroElement::MacroElement(int dim)
    : element(NULL),
      coord(dim, NO_INIT),
      boundary(dim, DEFAULT_VALUE, INTERIOR),
      projection(dim, NO_INIT),
      neighbour(dim, NO_INIT),
      oppVertex(dim, NO_INIT),
      index(-1), 
      elType(0),
      deserializedNeighbourIndices_(NULL)
  {
    neighbour.set(NULL);
    projection.set(NULL);
  }

  MacroElement::~MacroElement()
  {
    if(element) DELETE element;
  };

  void MacroElement::serialize(::std::ostream &out)
  {
    // write element-tree
    out << element->getTypeName() << ::std::endl;
    element->serialize(out);

    // write coords
    int i, size = coord.getSize();
    out.write(reinterpret_cast<const char*>(&size), sizeof(int));
    for(i = 0; i < size; i++) {
      coord[i].serialize(out);
    }

    // write boundary
    boundary.serialize(out);

    // write projection
    size = projection.getSize();
    out.write(reinterpret_cast<const char*>(&size), sizeof(int));
    for(i = 0; i < size; i++) {
      int id = projection[i] ? projection[i]->getID() : -1;
      out.write(reinterpret_cast<const char*>(&id), sizeof(int));
    }

    // write neighbour
    size = neighbour.getSize();
    out.write(reinterpret_cast<const char*>(&size), sizeof(int));
    for(i = 0; i < size; i++) {
      int index = neighbour[i] ? neighbour[i]->getIndex() : -1;
      out.write(reinterpret_cast<const char*>(&index), sizeof(int));
    }
  
    // write oppVertex
    oppVertex.serialize(out);

    // write index
    out.write(reinterpret_cast<const char*>(&index), sizeof(int));

    // write elType
    out.write(reinterpret_cast<const char*>(&elType), sizeof(unsigned char));
  }

  void MacroElement::deserialize(::std::istream &in)
  {
    // read element-tree
    ::std::string typeName;
    in >> typeName;
    in.get();

    if(element) {
      TEST_EXIT(typeName == element->getTypeName())("wrong element type name\n");
    } else {
      if(typeName == "Line") element = NEW Line(NULL);
      if(typeName == "Triangle") element = NEW Triangle(NULL);
      if(typeName == "Tetrahedron") element = NEW Tetrahedron(NULL);
    }

    element->deserialize(in);

    // read coords
    int i, size;
    in.read(reinterpret_cast<char*>(&size), sizeof(int));
    if(coord.getSize()) {
      TEST_EXIT(coord.getSize() == size)("invalid size\n");
    } else {
      coord.initSize(size);
    }
    for(i = 0; i < size; i++) {
      coord[i].deserialize(in);
    }

    // read boundary
    boundary.deserialize(in);

    // read projection
    in.read(reinterpret_cast<char*>(&size), sizeof(int));

    if(projection.getSize()) {
      TEST_EXIT(projection.getSize() == size)("invalid size\n");
    } else {
      projection.initSize(size);
    }

    for(i = 0; i < size; i++) {
      int id;
      in.read(reinterpret_cast<char*>(&id), sizeof(int));
      projection[i] = (id != -1) ? Projection::getProjection(id) : NULL;
    }

    // read neighbour indices
    in.read(reinterpret_cast<char*>(&size), sizeof(int));
  
    TEST_EXIT(deserializedNeighbourIndices_)
      ("neighbour indices for desrializing not set\n");
    
    deserializedNeighbourIndices_->resize(size);

    if(neighbour.getSize()) {
      TEST_EXIT(neighbour.getSize() == size)("invalid size\n");
    } else {
      neighbour.initSize(size);
    }

    for(i = 0; i < size; i++) {
      int index;
      in.read(reinterpret_cast<char*>(&index), sizeof(int));

      //     neighbour[i] = reinterpret_cast<MacroElement*>(index);
      //     // attention: the neighbour INDEX is stored in the neighbour vector.
      //     // After all macro elements are read in, the indices must be replaced
      //     // by the corresponding pointers.

      (*deserializedNeighbourIndices_)[i] = index;
    }

    deserializedNeighbourIndices_ = NULL;

    // read oppVertex
    oppVertex.deserialize(in);

    // write index
    in.read(reinterpret_cast<char*>(&index), sizeof(int));


    // write elType
    in.read(reinterpret_cast<char*>(&elType), sizeof(unsigned char));
  }

}
