// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file ParMetisPartitioner.h */

#ifndef AMDIS_PARMETISADAPTER_H
#define AMDIS_PARMETISADAPTER_H

#include "parmetis.h"
#include <map>
#include <set>
#include "MemoryManager.h"

namespace AMDiS {

  class FiniteElemSpace;

  enum PartitionMode {
    INITIAL = 0,          // initial partitioning of a unpartitioned mesh 
    ADAPTIVE_REPART = 1,  // repartitioning of a adaptively refined mesh
    REFINE_PART = 2       // quality improvement of the current partitioning
  };

  class ParMetisGraph;
  class Mesh;
  template<typename T > class DOFVector;

  class ParMetisMesh
  {
  public:
    MEMORY_MANAGED(ParMetisMesh);

    ParMetisMesh(Mesh *mesh);

    ~ParMetisMesh();

    inline void setParMetisIndex(int amdisIndex,
				 int parMetisIndex) 
    {
//       if(MPI::COMM_WORLD.Get_rank() == 0)
// 	MSG("set %p %d %d\n", this, amdisIndex, parMetisIndex);
      elem_a2p_[amdisIndex] = parMetisIndex + 1;
    };

    inline int getParMetisIndex(int amdisIndex) {
      int result = elem_a2p_[amdisIndex];
//       if(MPI::COMM_WORLD.Get_rank() == 0)
// 	MSG("get %p %d %d\n", this, amdisIndex, result - 1);
      TEST_EXIT(result > 0)("invalid index\n");
      return result - 1;
    };

    inline void setAMDiSIndex(int parMetisIndex,
			      int amdisIndex) 
    {
      elem_p2a_[parMetisIndex] = amdisIndex;
    };

    inline int getAMDiSIndex(int parMetisIndex) {
      return elem_p2a_[parMetisIndex];
    };

    inline int *getAMDiSIndices() {
      return elem_p2a_;
    };

    inline int *getElementPtr() { return eptr_; };
    inline int *getElementInd() { return eind_; };
    inline int *getElementDist() { return elmdist_; };
    inline int getDim() { return dim_; };
    inline float *getXYZ() { return xyz_; };
    inline int getNumElements() { return numElements_; };

  protected:
    int *eptr_;
    int *eind_;
    int *elmdist_;
    int dim_;
    float *xyz_;

    int numElements_;

    std::map<int, int> elem_a2p_;
    int *elem_p2a_;
  };

  class ParMetisGraph
  {
  public:
    MEMORY_MANAGED(ParMetisGraph);

    ParMetisGraph(ParMetisMesh *parMetisMesh,
		  int ncommonnodes = -1);

    ~ParMetisGraph();

    inline int *getXAdj() { return xadj_; };
    inline int *getAdjncy() { return adjncy_; };

  protected:
    ParMetisMesh *parMetisMesh_;

    int *xadj_;
    int *adjncy_;
  };

  class ParMetisPartitioner
  {
  public:
    ParMetisPartitioner(Mesh *mesh)
      : mesh_(mesh),
	parMetisMesh_(NULL)
    {};

    void partition(std::map<int, double> *elemWeights,
		   PartitionMode mode = INITIAL,
		   float itr = 1000000.0);

    void fillCoarsePartitionVec(std::map<int, int> *partitionVec);

    void fillLeafPartitionVec(std::map<int, int> *coarseVec,
			      std::map<int, int> *fineVec);

    void createPartitionData();

    void deletePartitionData();

  protected:
    void distributePartitioning(int *part);

    void descendPartitionData(Element *element);

  protected:
    Mesh *mesh_;
    ParMetisMesh *parMetisMesh_;
  };
}

#endif
