// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file Quadrature.h */

#ifndef AMDIS_QUADRATURE_H
#define AMDIS_QUADRATURE_H

#include "BasisFunction.h"
#include "Flag.h"
#include "MemoryManager.h"
#include "FixVec.h"
#include <list>

namespace AMDiS {

  template<typename T> class WorldVector;
  template<typename T> class DimVec;
  template<typename T> class VectorOfFixVecs;
  template<typename T> class MatrixOfFixVecs;

  // ============================================================================
  // ===== class Quadrature =====================================================
  // ============================================================================

  /** 
   * \ingroup Assembler
   *
   * \brief
   * For the assemblage of the system matrix and right hand side vector of the 
   * linear system, we have to compute integrals, for example: 
   * \f[ \int_{\Omega} f(x)\varphi_i(x) dx \f]
   * For general data A, b, c, and f, these integrals can not be calculated 
   * exactly. Quadrature formulas have to be used in order to calculate the 
   * integrals approximately. Numerical integration in finite element methods is
   * done by looping over all grid elements and using a quadrature formula on 
   * each element.
   */
  class Quadrature
  {
  public:
    MEMORY_MANAGED(Quadrature);

  protected:
    /** \brief
     * Avoids call of default constructor
     */
    Quadrature();

    /** \brief
     * Constructs a Quadrature with name name_ of degree degree_ for dim dim_.
     * The Quadrature has n_points_ quadrature points with barycentric 
     * coordinates lambda_ and weights w_. The constructor is protected because
     * access to a Quadrature should be done via \ref provideQuadrature.
     */
    Quadrature(const char* name_,
	       int         degree_,
	       int         dim_,
	       int         n_points_,
	       VectorOfFixVecs<DimVec<double> > *lambda_,
	       double*     w_)
      : name(name_),
	degree(degree_),
	dim(dim_),
	n_points(n_points_),
	lambda(lambda_),
	w(w_) 
    {};

    /** \brief
     * Destructor
     */
    ~Quadrature();

  public:
    /** \brief
     * Copy constructor
     */
    Quadrature(const Quadrature&);

    /** \brief
     * Returns a Quadrature for dimension dim_ exact for degree degree_.
     */
    static Quadrature *provideQuadrature(int dim_, int degree_);

    /** \brief
     * Approximates an integral by the numerical quadrature described by quad;
     * f is a pointer to an AbstractFunction to be integrated, evaluated in 
     * barycentric coordinates; the return value is 
     * \f[ \sum_{k = 0}^{n_points-1} w[k] * (*f)(lambda[k]) \f]
     * For the approximation of \f$ \int_S f\f$ we have to multiply this value 
     * with d!|S| for a simplex S; for a parametric simplex f should be a pointer
     * to a function which calculates 
     * \f$ f(\lambda)|det DF_S(\hat{x}(\lambda))| \f$.
     */
    double integrateStdSimplex(AbstractFunction<double, DimVec<double> > *f);
  
    /** \brief
     * Returns \ref name
     */
    inline const ::std::string& getName() { return name; };

    /** \brief
     * Returns \ref n_points
     */
    inline int getNumPoints() const { return n_points;};

    /** \brief
     * Returns \ref w[p]
     */
    inline double getWeight(int p) const {return w[p];};

    /** \brief
     * Returns \ref w.
     */
    inline double* getWeight() const { return w; };

    /** \brief
     * Returns \ref dim
     */
    inline int getDim() const { return dim; };

    /** \brief
     * Returns \ref degree
     */
    inline int getDegree() const { return degree; };

    /** \brief
     * Returns a pointer to a vector storing the values of a doubled valued 
     * function at all quadrature points; f is that AbstractFunction
     * , evaluated in barycentric coordinates; if vec is not NULL, the values are
     * stored in this vector, otherwise the values are stored in some static 
     * local vector, which is overwritten on the next call
     */
    const double *fAtQp(const AbstractFunction<double, DimVec<double> >& f,
			double *vec) const ;

    /** \brief
     * Returns a pointer to a vector storing the gradient (with respect to world
     * coordinates) of a double valued function at all quadrature points;
     * grdF is a pointer to a AbstractFunction, evaluated in barycentric
     * coordinates and returning a pointer to a WorldVector storing the gradient;
     * if vec is not NULL, the values are stored in this vector, otherwise the 
     * values are stored in some local static vector, which is overwritten on the
     * next call
     */
    const WorldVector<double> *grdFAtQp(const AbstractFunction<WorldVector<double>, 
					DimVec<double> >& grdF, 
					WorldVector<double>* vec) const;
  


    /** \brief
     * Returns \ref lambda[a][b] which is the b-th coordinate entry of the a-th
     * quadrature point
     */
    inline double getLambda(int a,int b) const {
      return (lambda ? (*lambda)[a][b] : 0.0);
    };

    /** \brief
     * Returns \ref lambda[a] which is a DimVec<double> containing the 
     * coordiantes of the a-th quadrature point
     */
    inline const DimVec<double>& getLambda(int a) const {
      return (*lambda)[a]; 
    };

    /** \brief
     * Returns \ref lambda which is a VectorOfFixvecs<DimVec<double> >.
     */
    VectorOfFixVecs<DimVec<double> > *getLambda() const { return lambda; };

    /** \brief
     * The function returns a pointer ptr to a vector of length 
     * \ref n_points storing the values of \f$ u_h \f$ at all 
     * quadrature points, i.e. 
     * \f[ ptr[l] = u_h(lambda[l]) \f] where 
     * l = 0, . . . , n_points - 1. vec is an optional memory pointer
     */
    //   const double *uhAtQp(const BasisFunction *basFcts,
    // 		       const double *uh_loc, double *vec) const;


    /** \brief
     * The function returns a pointer ptr to a vector of length 
     * \ref n_points WorldVectors storing \f$ \nabla u_h \f$ at all 
     * quadrature points, i.e.
     * \f[ ptr[l][i] = u_{h,xi}(lambda[l]) \f]
     * where l = 0, ... , n_points - 1, and i = 0, ... , 
     * DIM_OF_WORLD - 1; vec is an optional memory pointer
     */
    //   const WorldVector<double> *grdUhAtQp(const BasisFunction *basFcts,
    // 				       const DimVec<WorldVector<double> >& Lambda, 
    // 				       const double *uh_loc, 
    // 				       WorldVector<double> *vec) const;

    /** \brief
     * The function returns a pointer ptr to a vector of length 
     * \ref n_points of WorldMatrices storing D2uh at all quadrature 
     * points, i.e.
     * \f[ ptr[l][i][j] = u_{h,x_ix_j}(lambda[l]) \f]
     * where l = 0, ... , n_points - 1, and i, j = 0, ... , 
     * DIM_OF_WORLD - 1; vec is an optional memory pointer
     */
    //   const WorldMatrix<double> *D2UhAtQp(const BasisFunction *basFcts,
    // 				      const DimVec<WorldVector<double> >& Lambda, 
    // 				      const double *uh_loc, 
    // 				      WorldMatrix<double> *vec) const;
  

  public:
    /** \brief
     * Maximal number of quadrature points for the different dimensions
     */
    static const int maxNQuadPoints[4];

  protected:
    /** \brief
     * Name of this Quadrature
     */
    ::std::string name;

    /** \brief
     * Quadrature is exact of this degree
     */
    int degree;

    /** \brief
     * Quadrature for dimension dim
     */
    int dim;

    /** \brief
     * Number of quadrature points
     */
    int n_points;

    /** \brief
     * Vector of quadrature points given in barycentric coordinates
     */
    VectorOfFixVecs<DimVec<double> > *lambda; 

    /** \brief
     * Vector of quadrature weights
     */
    double *w;

  protected:
    /** \brief
     * Initialisation of all static Quadrature objects which will be returned
     * by \ref provideQuadrature()
     */
    static void initStaticQuadratures();

    /** \name static quadratures, used weights, and barycentric coords
     * \{
     */
    static Quadrature **quad_nd[4];
    static Quadrature *quad_0d[1];
    static Quadrature *quad_1d[20];
    static Quadrature *quad_2d[18];
    static Quadrature *quad_3d[8];

    static VectorOfFixVecs<DimVec<double> > *x_0d;
    static double *w_0d;

    static VectorOfFixVecs<DimVec<double> > *x0_1d;
    static VectorOfFixVecs<DimVec<double> > *x1_1d;
    static VectorOfFixVecs<DimVec<double> > *x2_1d;
    static VectorOfFixVecs<DimVec<double> > *x3_1d;
    static VectorOfFixVecs<DimVec<double> > *x4_1d;
    static VectorOfFixVecs<DimVec<double> > *x5_1d;
    static VectorOfFixVecs<DimVec<double> > *x6_1d;
    static VectorOfFixVecs<DimVec<double> > *x7_1d;
    static VectorOfFixVecs<DimVec<double> > *x8_1d;
    static VectorOfFixVecs<DimVec<double> > *x9_1d;
    static double *w0_1d;
    static double *w1_1d;
    static double *w2_1d;
    static double *w3_1d;
    static double *w4_1d;
    static double *w5_1d;
    static double *w6_1d;
    static double *w7_1d;
    static double *w8_1d;
    static double *w9_1d;

    static VectorOfFixVecs<DimVec<double> > *x1_2d;
    static VectorOfFixVecs<DimVec<double> > *x2_2d;
    static VectorOfFixVecs<DimVec<double> > *x3_2d;
    static VectorOfFixVecs<DimVec<double> > *x4_2d;
    static VectorOfFixVecs<DimVec<double> > *x5_2d;
    static VectorOfFixVecs<DimVec<double> > *x7_2d;
    static VectorOfFixVecs<DimVec<double> > *x8_2d;
    static VectorOfFixVecs<DimVec<double> > *x9_2d;
    static VectorOfFixVecs<DimVec<double> > *x10_2d;
    static VectorOfFixVecs<DimVec<double> > *x11_2d;
    static VectorOfFixVecs<DimVec<double> > *x12_2d;
    static VectorOfFixVecs<DimVec<double> > *x17_2d;
    static double *w1_2d;
    static double *w2_2d;
    static double *w3_2d;
    static double *w4_2d;
    static double *w5_2d;
    static double *w7_2d;
    static double *w8_2d;
    static double *w9_2d;
    static double *w10_2d;
    static double *w11_2d;
    static double *w12_2d;
    static double *w17_2d;

    static VectorOfFixVecs<DimVec<double> > *x1_3d;
    static VectorOfFixVecs<DimVec<double> > *x2_3d;
    static VectorOfFixVecs<DimVec<double> > *x3_3d;
    static VectorOfFixVecs<DimVec<double> > *x4_3d;
    static VectorOfFixVecs<DimVec<double> > *x5_3d;
    static VectorOfFixVecs<DimVec<double> > *x7_3d;
    static double *w1_3d;
    static double *w2_3d;
    static double *w3_3d;
    static double *w4_3d;
    static double *w5_3d;
    static double *w7_3d;

    /** \} */
  };

  // ============================================================================
  // ===== class FastQuadrature =================================================
  // ============================================================================

  /** \brief
   * Pre-compute the values of all basis functions at all quadrature nodes;  
   */
  const Flag INIT_PHI=1;

  /** \brief
   * Pre-compute the gradients (with respect to the barycentric coordinates) of
   * all basis functions at all quadrature nodes
   */
  const Flag INIT_GRD_PHI=2; 

  /** \brief
   * pre-compute all 2nd derivatives (with respect to the barycentric 
   * coordinates) of all basis functions at all quadrature nodes;
   */
  const Flag INIT_D2_PHI=4;

  // ============================================================================

  /** 
   * \ingroup Integration
   *
   *\brief
   * Often numerical integration involves basis functions, such as the assembling
   * of the system matrix and right hand side, or the integration of finite 
   * element functions. Since numerical quadrature involves only the values at 
   * the quadrature points and the values of basis functions and its derivatives
   * are the same at these points for all elements of the grid, such routines can
   * be much more efficient, if they can use pre-computed values of the basis
   * functions at the quadrature points. In this case the basis functions do not 
   * have to be evaluated for each quadrature point on every element newly.
   * Information that should be pre-computed can be specified by the following 
   * symbolic constants:
   * \ref INIT_PHI, \ref INIT_GRD_PHI, \ref INIT_D2_PHI
   */
  class FastQuadrature
  {
  public:
    MEMORY_MANAGED(FastQuadrature);

  protected:
    /** \brief
     * Constructs a FastQuadrature for the given Quadrature, BasisFunction, and
     * flag.
     */
    FastQuadrature(BasisFunction* basFcts, 
		   Quadrature* quad, 
		   Flag flag)
      : init_flag(flag), 
	phi(NULL), 
	grdPhi(NULL), 
	D2Phi(NULL), 
	quadrature(quad), 
	basisFunctions(basFcts) 
    {};

    /** \brief
     * Copy constructor
     */
    FastQuadrature(const FastQuadrature&);

    /** \brief
     * Extended copy constructor
     */
    FastQuadrature(const FastQuadrature&,const Flag);

    /** \brief
     * Destructor
     */
    ~FastQuadrature();

  public:
    /** \brief
     * Returns a FastQuadrature for the given BasisFunction, Quadrature, and
     * flags
     */
    static FastQuadrature* provideFastQuadrature(const BasisFunction*,
						 const Quadrature&,
						 Flag);

    /** \brief
     * inits FastQuadrature like speciefied in flag
     */
    void init(Flag init_flag);

    inline bool initialized(Flag flag) {
      if(flag == INIT_PHI) {
	return (phi != NULL);
      }

      if(flag == INIT_GRD_PHI) {
	return (grdPhi != NULL);
      }

      if(flag == INIT_D2_PHI) {
	return (D2Phi != NULL);
      }

      ERROR_EXIT("invalid flag\n");
      return false;
    };

    /** \brief
     * Returns \ref quadrature
     */
    inline const Quadrature* getQuadrature() const { return quadrature; };

    /** \brief
     * Returns \ref max_points
     */
    inline int getMaxQuadPoints() { return max_points; };

    /** \brief
     * Returns (*\ref D2Phi)[q][i][j][m]
     */
    const double getSecDer(int q,int i ,int j, int m) const;

    /** \brief
     * Returns (*\ref D2Phi)[q]
     */
    const VectorOfFixVecs<DimMat<double> > *getSecDer(int q) const;

    /** \brief
     * Returns (*\ref grdPhi)[q][i][j]
     */
    const double getGradient(int q, int i ,int j) const;

    /** \brief
     * Returns (*\ref grdPhi)[q]
     */
    VectorOfFixVecs<DimVec<double> >* getGradient(int q) const;

    /** \brief
     * Returns \ref phi[q][i]
     */
    inline const double getPhi(int q,int i) const {return (phi)?phi[q][i]:0;};

    /** \brief
     * Returns \ref phi[q]
     */
    inline const double *getPhi(int q) const {return (phi)?phi[q]:0;};

    /** \brief
     * Returns \ref quadrature ->integrateStdSimplex(f)
     */
    inline double 
    integrateStdSimplex(AbstractFunction<double, DimVec<double> > *f) {
      return quadrature->integrateStdSimplex(f); 
    };
  
    /** \brief
     * Returns \ref quadrature ->getNumPoints()
     */
    inline int getNumPoints() const { return quadrature->getNumPoints();};

    /** \brief
     * Returns \ref quadrature ->getWeight(p)
     */
    inline double getWeight(int p) const {return quadrature->getWeight(p);};

    /** \brief
     * Returns \ref quadrature ->getDim()
     */
    inline int getDim() const { return quadrature->getDim(); };

    /** \brief
     * Returns \ref quadrature ->getDegree()
     */
    inline int getDegree() const { return quadrature->getDegree(); };

    /** \brief
     * Returns \ref quadrature ->grdFAtQp(f, vec)
     */
    inline const WorldVector<double> 
    *grdFAtQp(const AbstractFunction<WorldVector<double>, 
	      DimVec<double> >& f,
	      WorldVector<double>* vec) const 
    { 
      return quadrature->grdFAtQp(f, vec);
    };
  
    /** \brief
     * Returns \ref quadrature ->fAtQp(f, vec)
     */
    inline const double *fAtQp(const AbstractFunction<double, 
			       DimVec<double> >& f,double *vec) const 
    {
      return quadrature->fAtQp(f, vec);    
    };

    /** \brief
     * Returns \ref quadrature ->getLambda(a,b)
     */
    inline double getLambda(int a,int b) const {
      return quadrature->getLambda(a,b);
    };

    /** \brief
     * Returns \ref quadrature ->getLambda(a)
     */
    inline const DimVec<double>& getLambda(int a) const { 
      return quadrature->getLambda(a); 
    };

    /** \brief
     * Returns \ref basisFunctions
     */
    inline BasisFunction* getBasisFunctions() const { return basisFunctions; };

    /** \brief
     * The function returns a pointer ptr to a vector of length 
     * \ref quadrature->n_points storing the values of \f$ u_h \f$ at all 
     * quadrature points of \ref quadrature, i.e. 
     * \f[ ptr[l] = u_h(quadrature->lambda[l]) \f] where 
     * l = 0, . . . , quadrature->n_points - 1; the \ref INIT_PHI flag must 
     * be set in \ref init_flag; vec is an optional memory pointer
     */
    //   const double *uhAtQp(const double *uh_loc, double *vec) const;

    /** \brief
     * The function returns a pointer ptr to a vector of length 
     * quadrature->n_points WorldVectors storing \f$ \nabla u_h \f$ at all 
     * quadrature points of \ref quadrature, i.e.
     * \f[ ptr[l][i] = u_{h,xi}(quadrature->lambda[l]) \f]
     * where l = 0, ... , quadrature->n_points - 1, and i = 0, ... , 
     * DIM_OF_WORLD - 1; the \ref INIT_GRD_PHI flag must be set in 
     * \ref init flag; vec is an optional memory pointer
     */
    //   const WorldVector<double> *grdUhAtQp(const DimVec<WorldVector<double> >& Lambda, 
    //   				       const double *uh_loc, 
    //   				       WorldVector<double> *vec) const;


    /** \brief
     * The function returns a pointer ptr to a vector of length 
     * quadrarure->n_points of WorldMatrices storing D2uh at all quadrature 
     * points of \ref quadrature, i.e.
     * \f[ ptr[l][i][j] = u_{h,x_ix_j}(quadrature->lambda[l]) \f]
     * where l = 0, ... , quadrature->n_points - 1, and i, j = 0, ... , 
     * DIM_OF_WORLD - 1; the \ref INIT_D2_PHI flag must be set in 
     * \ref init flag; vec is an optional memory pointer
     */
    //   const WorldMatrix<double> *D2UhAtQp(const DimVec<WorldVector<double> >& Lambda, 
    //   				      const double *uh_loc, 
    //   				      WorldMatrix<double> *vec) const;

  protected:
    /** \brief
     * Specifies which information should be pre-computed. Can be \ref INIT_PHI, 
     * \ref INIT_GRD_PHI, or \ref INIT_D2_PHI
     */
    Flag init_flag;

    /** \brief
     * Matrix storing function values if the flag \ref INIT_PHI is set;
     * phi[i][j] stores the value \ref basisFunctions->phi[j]
     * (quadrature->lambda[i]), 0 <= j < basisFunctions->getNumber()  and 
     * 0 <= i < n_points
     */
    double **phi;

    /** \brief
     * Matrix storing all gradients (with respect to the barycentric coordinates)
     * if the flag \ref INIT_GRD_PHI is set; grdPhi[i][j][k] stores the value 
     * basisFunctions->grdPhi[j](quadrature->lambda[i])[k] 
     * for 0 <= j < basisFunctions->getNumber(),
     * 0 <= i < . . . , n_points, and 0 <= k < DIM
     */
    MatrixOfFixVecs<DimVec<double> > *grdPhi;

    /** \brief
     * Matrix storing all second derivatives (with respect to the barycentric
     * coordinates) if the flag \ref INIT_D2_PHI is set; D2Phi[i][j][k][l] stores
     * the value basisFunctions->D2Phi[j](quadrature->lambda[i])[k][l] 
     * for 0 <= j < basisFunctions->getNumber(), 
     * 0 <= i < n_points, and 0 <= k,l < DIM
     */
    MatrixOfFixVecs<DimMat<double> > *D2Phi;

    /** \brief
     * List of all used FastQuadratures
     */
    static ::std::list<FastQuadrature*> fastQuadList;

    /** \brief
     * Maximal number of quadrature points for all yet initialised FastQuadrature
     * objects. This value may change after a new initialisation of a
     * FastQuadrature
     */
    static int max_points;

    /** \brief
     * This FastQuadrature stores values for Quadrature quadrature
     */
    Quadrature* quadrature;

    /** \brief
     * Values stored for basis functions basisFunctions
     */
    BasisFunction* basisFunctions;

  };

}

#endif  // AMDIS_QUADRATURE_H 
