#include "AMDiS.h"
#include "mpi.h"
#include "ParallelProblem.h"

using namespace std;
using namespace AMDiS;

// ===== function definitions // 
/** \brief
 * Dirichlet boundary function
 */
class G : public AbstractFunction<double, WorldVector<double> >
{
public:
  MEMORY_MANAGED(G);

  /** \brief
   * Implementation of AbstractFunction::operator().
   */
  const double& operator()(const WorldVector<double>& x) const
  {
    static double result;
    result = exp(-10.0*(x*x));
    return result;  
  };
};

class GrdG : public AbstractFunction<WorldVector<double>, WorldVector<double> >
{
  MEMORY_MANAGED(GrdG);

  /** \brief
   * Implementation of AbstractFunction::operator().
   */
  const WorldVector<double>& operator()(const WorldVector<double>& x) const
  {
    static WorldVector<double> result;
    result = x * -20.0 * exp(-10.0*(x*x));
    return result;  
  };
};

/** \brief
 * RHS function
 */
class F : public AbstractFunction<double, WorldVector<double> >
{
public:
  MEMORY_MANAGED(F);

  F(int degree) : AbstractFunction<double, WorldVector<double> >(degree) {};

  /** \brief
   * Implementation of AbstractFunction::operator().
   */
  const double& operator()(const WorldVector<double>& x) const {
    static double result = 0.0;
    int dim = x.getSize();
    double r2 = (x*x);
    double ux = exp(-10.0*r2);
    result = -(400.0*r2 - 20.0*dim)*ux;
    return result;
  };
};

// ===== main program ===== 
int main(int argc, char* argv[])
{
  MPI::Init(argc, argv);

  FUNCNAME("main");

  // ===== check for init file =====
  TEST_EXIT(argc >= 2)("usage: ellipt initfile\n");

  // ===== init parameters =====
  Parameters::init(false, argv[1]);

  // ===== create and init the scalar problem ===== 
  ProblemScal ellipt("ellipt");
  ellipt.initialize(INIT_ALL);

  // === parallel problem ===
  std::vector<DOFVector<double>*> vectors;
  ParallelProblemScal parallelellipt("ellipt->parallel", &ellipt, NULL, vectors);

  // === create adapt info ===
  AdaptInfo *adaptInfo = NEW AdaptInfo("ellipt->adapt", 1);

  // === create adapt ===
  AdaptStationary *adapt = NEW AdaptStationary("ellipt->adapt",
					       &parallelellipt,
					       adaptInfo);
  
  // ===== add boundary conditions =====
  ellipt.addDirichletBC(1, NEW G);

  // ===== create matrix operator =====
  Operator matrixOperator(Operator::MATRIX_OPERATOR, ellipt.getFESpace());
  matrixOperator.addSecondOrderTerm(NEW Laplace_SOT);
  ellipt.addMatrixOperator(&matrixOperator);

  // ===== create rhs operator =====
  int degree = ellipt.getFESpace()->getBasisFcts()->getDegree();
  Operator rhsOperator(Operator::VECTOR_OPERATOR, ellipt.getFESpace());
  rhsOperator.addZeroOrderTerm(NEW CoordsAtQP_ZOT(NEW F(degree)));
  ellipt.addVectorOperator(&rhsOperator);

  parallelellipt.initParallelization(adaptInfo);
  adapt->adapt();
  parallelellipt.exitParallelization(adaptInfo);

  MPI::Finalize();
}


